# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import copy
import unittest

from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.system.architecture_entry import ArchitectureEntry
from availsim4core.src.context.system.children_logic.and_ import And
from availsim4core.src.context.system.children_logic.children_logic import ChildrenLogic
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.component_tree.component_factory import ComponentFactory
from availsim4core.src.context.system.component_tree.compound import Compound
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.failure import Failure, FailureType
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.context.system.failure_mode_assignments import FailureModeAssignments
from availsim4core.src.context.system.inspection import Inspection
from availsim4core.src.context.system.minimal_replaceable_unit import MinimalReplaceableUnit
from availsim4core.src.context.system.probability_law.probability_law import ProbabilityLaw
from availsim4core.src.context.system import sanity_check
from availsim4core.src.context.system.system_template import SystemTemplate
from availsim4core.src.reader.xlsx.system_template_reader import SystemTemplateSheet, SystemTemplateReader


class test_ComponentFactory(unittest.TestCase):

    def test_build(self):
        root_name = "ROOT_NODE"
        basic_name = "test"
        expected_mru = MinimalReplaceableUnit("test_mru",
                                              ProbabilityLaw("", 0, False),
                                              "IMMEDIATE",
                                              Status.FAILED,
                                              root_name)

        NO_PHASE = Phase("NONE", ProbabilityLaw("", 0, False), False)
        NO_PHASE.set_next_phase(NO_PHASE)
        NO_PHASE.set_next_phase_if_failure(NO_PHASE)

        expected_inspection = Inspection("inspection_test", 1, 2)

        failure_mode_1 = FailureMode("TEST_FAILURE",
                                     ProbabilityLaw("", 0, False),
                                     ProbabilityLaw("", 0, False),
                                     Failure(FailureType.DETECTABLE),
                                     {NO_PHASE},
                                     expected_inspection,
                                     {NO_PHASE},
                                     NO_PHASE,
                                     'AFTER_REPAIR',
                                     {NO_PHASE})

        failure_mode_assignment_1 = FailureModeAssignments(basic_name, failure_mode_1)
        expected_root_node = Compound(1, root_name, 0, [], And(), [])
        expected_basic_1 = Basic(2,
                                 basic_name,
                                 0,
                                 [expected_root_node],
                                 failure_mode_1,
                                 [expected_mru],
                                 [expected_mru])

        architecture_entry_root = ArchitectureEntry(root_name, "COMPOUND", 1, [basic_name], "AND", [], [])
        architecture_entry_basic = ArchitectureEntry(basic_name,
                                                     "BASIC", 1, [], "",
                                                     ["test_mru"], ["test_mru"])
        system_template = SystemTemplate([architecture_entry_root, architecture_entry_basic],
                                         [failure_mode_assignment_1],
                                         [failure_mode_1],
                                         [expected_mru],
                                         [expected_inspection],
                                         {NO_PHASE},
                                         set(),
                                         set(),
                                         None)

        result_root_node = ComponentFactory(system_template).build()

        self.assertEqual(result_root_node, expected_root_node)
        result_basic = result_root_node._children[0]
        self.assertEqual(result_basic, expected_basic_1)
        self.assertEqual(result_basic.failure_mode, expected_basic_1.failure_mode)
        self.assertEqual(result_basic.list_of_mru_trigger, expected_basic_1.list_of_mru_trigger)
        self.assertEqual(result_basic.list_of_mru_group, expected_basic_1.list_of_mru_group)

    def test_add_component(self):

        failure_mode_1 = FailureMode("ACCELERATOR_CONTROLS_FAILURE",
                                     ProbabilityLaw("", 0, False),
                                     ProbabilityLaw("", 0, False),
                                     Failure(FailureType.DETECTABLE),
                                     "[]",
                                     None,
                                     [],
                                     None,
                                     'AFTER_REPAIR',
                                     [])

        failure_mode_assignment_1 = FailureModeAssignments("ACCELERATOR_CONTROLS", [failure_mode_1])

        component_name: str = "ACCELERATOR_CONTROLS"

        architecture_entry_root = ArchitectureEntry("ROOT_NODE", "COMPOUND", 1, [component_name], "AND", [], [])
        architecture_entry_basic = ArchitectureEntry(component_name, "BASIC", 1, [], "", [], [])
        system_template = SystemTemplate([architecture_entry_root, architecture_entry_basic],
                                         [failure_mode_assignment_1],
                                         [failure_mode_1],
                                         [],
                                         [],
                                         None, 
                                         set(),
                                         set(),
                                         None)

        component_factory = ComponentFactory(system_template)

        result_list = component_factory._add_component(architecture_entry_root, [], None)

        expected_result = Compound(1, "LINAC4", 0, [], And(), [])
        expected_basic = Basic(2, "ACCELERATOR_CONTROLS", 0, [expected_result], failure_mode_1, [], [])
        expected_result.add_children_list([expected_basic])
        expected_result_list = [expected_result]

        for result in result_list:
            for expected_result in expected_result_list:
                self.assertEqual(result, expected_result)
                self.assertEqual(result._children, expected_result._children)

    def test_add_basics(self):

        parents = [Compound(0, "ROOT_NODE", 0, [], ChildrenLogic(), [])]

        failure_mode_1 = FailureMode("ACCELERATOR_CONTROLS_FAILURE",
                                     ProbabilityLaw("", 0, False),
                                     ProbabilityLaw("", 0, False),
                                     Failure(FailureType.DETECTABLE),
                                     "[]",
                                     None,
                                     [],
                                     None,
                                     'AFTER_REPAIR',
                                     [])
        failure_mode_assignment_1 = FailureModeAssignments("test", [failure_mode_1])

        expected_basic_1 = Basic(1, "test", 0, parents, failure_mode_1, [], [])
        expected_basic_2 = Basic(2, "test", 1, parents, failure_mode_1, [], [])
        expected_results = [expected_basic_1, expected_basic_2]

        architecture_entry_root = ArchitectureEntry("ROOT_NODE", "COMPOUND", 1, ["test"], "", [], [])
        architecture_entry_basic = ArchitectureEntry("test", "BASIC", 2, [], "", [], [])

        system_template = SystemTemplate([architecture_entry_root, architecture_entry_basic],
                                         [failure_mode_assignment_1],
                                         [failure_mode_1],
                                         [],
                                         [],
                                         None, 
                                         set(),
                                         set(),
                                         None)
        component_factory = ComponentFactory(system_template)

        result_basic_list = component_factory._add_basics(architecture_entry_basic,
                                                          parents)

        self.assertListEqual(expected_results, result_basic_list)

    def test_add_compound(self):
        failure_mode_1 = FailureMode("ACCELERATOR_CONTROLS_FAILURE",
                                     ProbabilityLaw("", 0, False),
                                     ProbabilityLaw("", 0, False),
                                     Failure(FailureType.DETECTABLE),
                                     "[]",
                                     None,
                                     [],
                                     None,
                                     'AFTER_REPAIR',
                                     [])

        failure_mode_assignment_1 = FailureModeAssignments("ACCELERATOR_CONTROLS", [failure_mode_1])

        architecture_entry_root = ArchitectureEntry("LINAC4", "COMPOUND", 1, ["ACCELERATOR_CONTROLS"], "AND", [], [])
        architecture_entry_basic = ArchitectureEntry("ACCELERATOR_CONTROLS", "BASIC", 1, [], "", [], [])

        system_template = SystemTemplate([architecture_entry_root, architecture_entry_basic],
                                         [failure_mode_assignment_1],
                                         [failure_mode_1],
                                         [],
                                         [],
                                         None, 
                                         set(),
                                         set(),
                                         None)
        component_factory = ComponentFactory(system_template)

        result_list = component_factory._add_compound(architecture_entry_root, [], None)

        expected_result = Compound(1, "LINAC4", 0, [], And(), [])
        expected_basic = Basic(2, "ACCELERATOR_CONTROLS", 0, [expected_result], failure_mode_1, [], [])
        expected_result.add_children_list([expected_basic])
        expected_result_list = [expected_result]

        for result in result_list:
            for expected_result in expected_result_list:
                self.assertEqual(result, expected_result)
                self.assertTrue(result, expected_result)

    def test_add_shared_children_compound(self):

        failure_mode_1 = FailureMode("test_shared_childFailure",
                                     ProbabilityLaw("", 0, False),
                                     ProbabilityLaw("", 0, False),
                                     Failure(FailureType.DETECTABLE),
                                     "[]",
                                     None,
                                     [],
                                     None,
                                     'AFTER_REPAIR',
                                     [])

        failure_mode_assignment_1 = FailureModeAssignments("test_shared_child", [failure_mode_1])

        architecture_entry_root = ArchitectureEntry("LINAC4", "COMPOUND", 1, ["test_compound_1", "test_compound_2"],
                                                    "AND", [], [])
        architecture_entry_compound_1 = ArchitectureEntry("test_compound_1", "COMPOUND", 1,
                                                          ["test_shared_child(LINAC4)"], "AND", [], [])
        architecture_entry_compound_2 = ArchitectureEntry("test_compound_2", "COMPOUND", 1,
                                                          ["test_shared_child(LINAC4)"], "AND", [], [])
        architecture_entry_basic = ArchitectureEntry("test_shared_child", "BASIC", 1, [], "", [], [])

        architecture_entry_list = [architecture_entry_root, architecture_entry_compound_1,
                                   architecture_entry_compound_2,
                                   architecture_entry_basic]

        system_template = SystemTemplate(architecture_entry_list,
                                         [failure_mode_assignment_1],
                                         [failure_mode_1],
                                         [],
                                         [],
                                         None, 
                                         set(),
                                         set(),
                                         None)
        component_factory = ComponentFactory(system_template)

        result_list = component_factory._add_compound(architecture_entry_root, [], None)

        expected_root_node = Compound(1, "LINAC4", 0, [], And(), [])
        expected_result_compound_1 = Compound(2, "test_compound_1", 0, [expected_root_node], And(), [])
        expected_result_compound_2 = Compound(3, "test_compound_2", 0, [expected_root_node], And(), [])
        expected_result_basic = Basic(4, "test_shared_child", 0,
                                      [expected_result_compound_1, expected_result_compound_2],
                                      [failure_mode_1], [], [])
        expected_result_compound_1.add_children_list([expected_result_basic])
        expected_result_compound_2.add_children_list([expected_result_basic])
        expected_root_node.add_children_list([expected_result_compound_1, expected_result_compound_2])

        expected_result_list = [expected_root_node]

        for result in result_list:
            for expected_result in expected_result_list:
                # TODO since we test only the uniq id, the full tree is not tested here as it should
                self.assertEqual(result, expected_result)

    def test_add_shared_children_compound__multiple_shared_children(self):

        failure_mode_1 = FailureMode("test_shared_childFailure",
                                     ProbabilityLaw("", 0, False),
                                     ProbabilityLaw("", 0, False),
                                     Failure(FailureType.DETECTABLE),
                                     "[]",
                                     None,
                                     [],
                                     None,
                                     'AFTER_REPAIR',
                                     [])

        failure_mode_assignment_1 = FailureModeAssignments("test_shared_child_1", [failure_mode_1])
        failure_mode_assignment_2 = FailureModeAssignments("test_shared_child_2", [failure_mode_1])

        architecture_entry_root = ArchitectureEntry("LINAC4", "COMPOUND", 1, ["test_compound_1", "test_compound_2"],
                                                    "AND", [], [])
        architecture_entry_compound_1 = ArchitectureEntry("test_compound_1", "COMPOUND", 1,
                                                          ["test_shared_child_1(LINAC4)",
                                                           "test_shared_child_2(LINAC4)"],
                                                          "AND", [], [])
        architecture_entry_compound_2 = ArchitectureEntry("test_compound_2", "COMPOUND", 1,
                                                          ["test_shared_child_1(LINAC4)",
                                                           "test_shared_child_2(LINAC4)"],
                                                          "AND", [], [])
        architecture_entry_shared_basic_1 = ArchitectureEntry("test_shared_child_1", "BASIC", 1, [], "", [], [])
        architecture_entry_shared_basic_2 = ArchitectureEntry("test_shared_child_2", "BASIC", 1, [], "", [], [])

        architecture_entry_list = [architecture_entry_root,
                                   architecture_entry_compound_1, architecture_entry_compound_2,
                                   architecture_entry_shared_basic_1, architecture_entry_shared_basic_2]

        system_template = SystemTemplate(architecture_entry_list,
                                         [failure_mode_assignment_1, failure_mode_assignment_2],
                                         [failure_mode_1],
                                         [],
                                         [],
                                         None, 
                                         set(),
                                         set(),
                                         None)
        component_factory = ComponentFactory(system_template)

        result_list = component_factory._add_compound(architecture_entry_root, [], None)

        expected_root_node = Compound(1, "LINAC4", 0, [], And(), [])
        expected_result_compound_1 = Compound(2, "test_compound_1", 0, [expected_root_node], And(), [])
        expected_result_compound_2 = Compound(5, "test_compound_2", 0, [expected_root_node], And(), [])
        expected_result_shared_basic_1 = Basic(3, "test_shared_child_1", 0,
                                               [expected_result_compound_1, expected_result_compound_2],
                                               [failure_mode_1], [], [])
        expected_result_shared_basic_2 = Basic(4, "test_shared_child_2", 0,
                                               [expected_result_compound_1, expected_result_compound_2],
                                               [failure_mode_1], [], [])

        result_root = result_list[0]
        self.assertEqual(result_root, expected_root_node)
        for root_child in result_root._children:
            self.assertEqual(root_child._children, [expected_result_shared_basic_1, expected_result_shared_basic_2])

    def test_add_shared_children_compound__many_instances_of_shared_children(self):

        failure_mode_1 = FailureMode("test_shared_childFailure",
                                     ProbabilityLaw("", 0, False),
                                     ProbabilityLaw("", 0, False),
                                     Failure(FailureType.DETECTABLE),
                                     "[]",
                                     None,
                                     [],
                                     None,
                                     'AFTER_REPAIR',
                                     [])

        failure_mode_assignment_1 = FailureModeAssignments("test_shared_child_1", [failure_mode_1])

        architecture_entry_root = ArchitectureEntry("LINAC4", "COMPOUND", 1, ["test_compound_1", "test_compound_2"],
                                                    "AND", [], [])
        architecture_entry_compound_1 = ArchitectureEntry("test_compound_1", "COMPOUND", 1,
                                                          ["test_shared_child_1(LINAC4)"],
                                                          "AND", [], [])
        architecture_entry_compound_2 = ArchitectureEntry("test_compound_2", "COMPOUND", 1,
                                                          ["test_shared_child_1(LINAC4)"],
                                                          "AND", [], [])
        architecture_entry_shared_basic_1 = ArchitectureEntry("test_shared_child_1", "BASIC", 3, [], "", [], [])

        architecture_entry_list = [architecture_entry_root,
                                   architecture_entry_compound_1, architecture_entry_compound_2,
                                   architecture_entry_shared_basic_1]

        system_template = SystemTemplate(architecture_entry_list,
                                         [failure_mode_assignment_1],
                                         [failure_mode_1],
                                         [],
                                         [],
                                         None, 
                                         set(),
                                         set(),
                                         None)
        component_factory = ComponentFactory(system_template)

        result_list = component_factory._add_compound(architecture_entry_root, [], None)

        expected_root_node = Compound(1, "LINAC4", 0, [], And(), [])
        expected_result_compound_1 = Compound(2, "test_compound_1", 0, [expected_root_node], And(), [])
        expected_result_compound_2 = Compound(6, "test_compound_2", 0, [expected_root_node], And(), [])
        expected_result_shared_basic_1 = Basic(3, "test_shared_child_1", 0,
                                               [expected_result_compound_1, expected_result_compound_2],
                                               [failure_mode_1], [], [])
        expected_result_shared_basic_2 = Basic(4, "test_shared_child_1", 0,
                                               [expected_result_compound_1, expected_result_compound_2],
                                               [failure_mode_1], [], [])
        expected_result_shared_basic_3 = Basic(5, "test_shared_child_1", 0,
                                               [expected_result_compound_1, expected_result_compound_2],
                                               [failure_mode_1], [], [])

        result_root = result_list[0]
        self.assertEqual(result_root, expected_root_node)
        for root_child in result_root._children:
            self.assertEqual(root_child._children, [expected_result_shared_basic_1,
                                                    expected_result_shared_basic_2,
                                                    expected_result_shared_basic_3])

    def test_find_shared_children(self):
        root_node = Compound(1, "ROOT_NODE", 0, [], ChildrenLogic(), [])
        parent_1 = Compound(2, "A", 1, [root_node], ChildrenLogic(), [])
        shared_child_compound = Compound(3, "C", 3, [parent_1], ChildrenLogic(), [])
        parent_2 = Compound(4, "B", 2, [root_node], ChildrenLogic(), [])

        shared_child_name_str = "C(ROOT_NODE)"
        shared_children_list = [shared_child_compound]

        result = ComponentFactory.find_shared_children(parent_2,
                                                       shared_child_name_str,
                                                       shared_children_list)
        self.assertEqual([shared_child_compound], result)

    def test_find_shared_children__multiple_shared_children(self):
        root_node = Compound(1, "ROOT_NODE", 0, [], ChildrenLogic(), [])
        parent_1 = Compound(2, "A", 1, [root_node], ChildrenLogic(), [])
        shared_child_compound0 = Compound(3, "C", 0, [parent_1], ChildrenLogic(), [])
        shared_child_compound1 = Compound(4, "C", 1, [parent_1], ChildrenLogic(), [])
        parent_2 = Compound(5, "B", 2, [root_node], ChildrenLogic(), [])

        shared_child_name_str = "C(ROOT_NODE)"
        shared_children_list = [shared_child_compound0, shared_child_compound1]

        result = ComponentFactory.find_shared_children(parent_2,
                                                       shared_child_name_str,
                                                       shared_children_list)
        self.assertEqual(result, [shared_child_compound0, shared_child_compound1])

    def test_compare_and_log_diff__lists_matching(self):

        list_one = ['A','A','A','A','B','C']
        list_two = ['A','B','C']

        matching = sanity_check._compare_and_log_diff(list_one, list_two, "test")

        self.assertEqual(matching,True)

    def test_compare_and_log_diff__lists_not_matching_v1(self):

        list_one = ['A','A','A','A','B','C']
        list_two = ['A','B','C','D']

        matching = sanity_check._compare_and_log_diff(list_one, list_two, "test")

        self.assertEqual(matching,False)


    def test_compare_and_log_diff__lists_not_matching_v2(self):
        list_one = ['A', 'A', 'A', 'A', 'B', 'C','D']
        list_two = ['A', 'B', 'C']

        matching = sanity_check._compare_and_log_diff(list_one, list_two, "test")

        self.assertEqual(matching,False)


    def test_sanity_check_of_the_tree(self):

        # defining a dictionary similar to a reader output

        default_dictionary = {
            "ARCHITECTURE":{
                0: {"COMPONENT_NAME": 'ROOT',
                    "COMPONENT_TYPE": 'COMPOUND',
                    "COMPONENT_NUMBER": 1,
                    "CHILDREN_NAME": 'ACCELERATORCONTROLS,ACCESSMANAGEMENT',####,BEAMLOSSES',
                    "CHILDREN_LOGIC": "AND",
                    "IN_MRU": 'none',
                    "TRIGGER_MRU": 'none'},
                1: {"COMPONENT_NAME": 'ACCELERATORCONTROLS',
                    "COMPONENT_TYPE": 'basic',
                    "COMPONENT_NUMBER": 2,
                    "CHILDREN_NAME": 'none',
                    "CHILDREN_LOGIC": 'none',
                    "IN_MRU": 'LHC_MRU',
                    "TRIGGER_MRU": 'none'},
                2: {"COMPONENT_NAME": 'ACCESSMANAGEMENT',
                    "COMPONENT_TYPE": 'basic',
                    "COMPONENT_NUMBER": 1,
                    "CHILDREN_NAME": 'none',
                    "CHILDREN_LOGIC": 'none',
                    "IN_MRU": 'none',
                    "TRIGGER_MRU": 'LHC_MRU'}
            },
            "FAILURE_MODE_ASSIGNMENTS": {
                0: {"COMPONENT_NAME": 'ACCELERATORCONTROLS', "FAILURE_MODE_NAME": 'ACCELERATORCONTROLSFAILURE'},
                1: {"COMPONENT_NAME": 'ACCESSMANAGEMENT', "FAILURE_MODE_NAME": 'ACCESSMANAGEMENTFAILURE'}
            },
            "FAILURE_MODES": {
                0: {"FAILURE_MODE_NAME": 'ACCELERATORCONTROLSFAILURE',
                    "FAILURE_LAW": "EXP",
                    "FAILURE_PARAMETERS": 300.0,
                    "REPAIR_LAW": "FIX",
                    "REPAIR_PARAMETERS": 1.5436,
                    "TYPE_OF_FAILURE": "DETECTABLE",
                    "HELD_BEFORE_REPAIR": "NEVER_HELD",
                    "INSPECTION_NAME": 'none',
                    "PHASE_NAME": 'OPERATION',
                    "NEXT_PHASE_IF_FAILURE": 'OPERATION',
                    "PHASE_CHANGE_TRIGGER": "AFTER_FAILURE",
                    "HELD_AFTER_REPAIR":"NEVER_HELD"},
                1: {"FAILURE_MODE_NAME": 'ACCESSMANAGEMENTFAILURE',
                    "FAILURE_LAW": "EXP",
                    "FAILURE_PARAMETERS": 859.74,
                    "TYPE_OF_FAILURE": "DETECTABLE",
                    "HELD_BEFORE_REPAIR": "NEVER_HELD",
                    "REPAIR_LAW": "FIX",
                    "REPAIR_PARAMETERS": 3.6167,
                    "INSPECTION_NAME": 'INSPECT',
                    "PHASE_NAME": 'OPERATION',
                    "NEXT_PHASE_IF_FAILURE": 'none',
                    "PHASE_CHANGE_TRIGGER": "AFTER_FAILURE",
                    "HELD_AFTER_REPAIR":"NEVER_HELD"},
            },
            "MRU":{
                0: {'MRU_NAME': 'LHC_MRU',
                    "MRU_LAW": "FIX",
                    "MRU_PARAMETERS": 0.001,
                    "MRU_SCHEDULE": "IMMEDIATE",
                    "LOWEST_COMMON_ANCESTOR_SCOPE": 'ROOT',
                    "TRIGGERING_STATUS": 'FAILED'}
            },
            "INSPECTIONS":{
                0: {"INSPECTION_NAME": 'INSPECT',
                    "INSPECTION_PERIOD":5,
                    "INSPECTION_DURATION": 2}
            },
            "PHASES":{
                0: {"PHASE_NAME": 'OPERATION',
                    "PHASE_LAW": 'FIX',
                    "PHASE_PARAMETERS": 1,
                    "FIRST_PHASE":"True",
                    "NEXT_DEFAULT_PHASE":'OPERATION',
                    "NEXT_DEFAULT_PHASE_IF_FAILURE":'OPERATION'
                    }
            }
        }

        def comparing_a_dictionary_to_its_tree(system_dictionary):

            # creating a system template

            architecture_entry_list = SystemTemplateReader.generate_architecture_entry_list(
                system_dictionary[SystemTemplateSheet.ARCHITECTURE])

            mru_list = SystemTemplateReader.generate_mrus(
                system_dictionary[SystemTemplateSheet.MINIMAL_REPLACEABLE_UNIT])

            inspections_list = SystemTemplateReader.generate_inspections(
                system_dictionary[SystemTemplateSheet.INSPECTIONS])

            phases_list = SystemTemplateReader.generate_phases(
                system_dictionary[SystemTemplateSheet.PHASES])

            failure_modes_list = SystemTemplateReader.generate_failure_modes(
                system_dictionary[SystemTemplateSheet.FAILURE_MODES],
                inspections_list,
                phases_list)

            failure_mode_assignments_list = SystemTemplateReader.generate_failure_mode_assignments(
                system_dictionary[SystemTemplateSheet.FAILURE_MODE_ASSIGNMENTS],
                failure_modes_list)

            system_template = SystemTemplate(architecture_entry_list,
                                             failure_mode_assignments_list,
                                             failure_modes_list,
                                             mru_list,
                                             inspections_list,
                                             phases_list,
                                             set(),
                                             set(),
                                             None)

            # building a system

            component_factory = ComponentFactory(system_template)
            component_list = component_factory._add_component(system_template.architecture_entry_list[0], [], None)

            matching = sanity_check.run(system_template, component_list[0])

            return matching

        matching = comparing_a_dictionary_to_its_tree(default_dictionary)
        self.assertEqual(matching,True)

        # modifying some entries but still matching dictionary and tree

        modified_dictionary = copy.deepcopy(default_dictionary)
        modified_dictionary['ARCHITECTURE'][0]["COMPONENT_NAME"]="root"
        matching = comparing_a_dictionary_to_its_tree(modified_dictionary)
        self.assertEqual(matching,True)

        modified_dictionary = copy.deepcopy(default_dictionary)
        modified_dictionary['ARCHITECTURE'][0]["CHILDREN_LOGIC"] = "and"
        matching = comparing_a_dictionary_to_its_tree(modified_dictionary)
        self.assertEqual(matching, True)

        modified_dictionary = copy.deepcopy(default_dictionary)
        modified_dictionary['ARCHITECTURE'][0]["CHILDREN_LOGIC"] = " and "
        matching = comparing_a_dictionary_to_its_tree(modified_dictionary)
        self.assertEqual(matching, True)

        modified_dictionary = copy.deepcopy(default_dictionary)
        modified_dictionary['FAILURE_MODES'][0]["FAILURE_MODE_NAME"]='AccELERATORCONTROLSFAILURE'
        matching = comparing_a_dictionary_to_its_tree(modified_dictionary)
        self.assertEqual(matching,True)

        modified_dictionary = copy.deepcopy(default_dictionary)
        modified_dictionary['FAILURE_MODES'][0]["FAILURE_MODE_NAME"]=' ACCELERATORCONTROLSFAILURE'
        matching = comparing_a_dictionary_to_its_tree(modified_dictionary)
        self.assertEqual(matching,True)

        modified_dictionary = copy.deepcopy(default_dictionary)
        modified_dictionary['ARCHITECTURE'][1]["COMPONENT_NUMBER"] = ' 2'
        matching = comparing_a_dictionary_to_its_tree(modified_dictionary)
        self.assertEqual(matching, True)

        modified_dictionary = copy.deepcopy(default_dictionary)
        modified_dictionary['ARCHITECTURE'][1]["COMPONENT_NUMBER"] = 2.
        matching = comparing_a_dictionary_to_its_tree(modified_dictionary)
        self.assertEqual(matching, True)

        # modifying some entries, dictionary and tree don't match

        modified_dictionary = copy.deepcopy(default_dictionary)
        modified_dictionary['ARCHITECTURE'][3] = \
            {"COMPONENT_NAME": 'DUMMY_COMPONENT_NOT_USED',
             "COMPONENT_TYPE": 'basic',
             "COMPONENT_NUMBER": 1,
             "CHILDREN_NAME": 'none',
             "CHILDREN_LOGIC": 'none',
             "IN_MRU": 'none',
             "TRIGGER_MRU": 'LHC_MRU'}
        matching = comparing_a_dictionary_to_its_tree(modified_dictionary)
        self.assertEqual(matching, False)

        modified_dictionary = copy.deepcopy(default_dictionary)
        modified_dictionary['FAILURE_MODE_ASSIGNMENTS'] = {
            0: {"COMPONENT_NAME": 'ACCELERATORCONTROLS', "FAILURE_MODE_NAME": 'ACCELERATORCONTROLSFAILURE'}
        }
        matching = comparing_a_dictionary_to_its_tree(modified_dictionary)
        self.assertEqual(matching, False)

        # modifying some entries, dictionary and tree don't match explicitly but still the test is passed
        # the phase added "DOWNTIME" is defined but never explicitly used in the input file, still it's present in the
        # tree because of the never_held option which uses every phases described

        modified_dictionary = copy.deepcopy(default_dictionary)
        modified_dictionary['PHASES'][1] = \
            {"PHASE_NAME": 'DOWNTIME',
             "PHASE_LAW": 'FIX',
             "PHASE_PARAMETERS": 1,
             "FIRST_PHASE": "True",
             "NEXT_DEFAULT_PHASE": 'OPERATION',
             "NEXT_DEFAULT_PHASE_IF_FAILURE": 'OPERATION'
             }
        matching = comparing_a_dictionary_to_its_tree(modified_dictionary)
        self.assertEqual(matching, True)


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest
from unittest.mock import patch

from availsim4core.src.reader.xlsx.sensitivity_analysis_reader import SensitivityAnalysisReader
from availsim4core.src.sensitivity_analysis.exploration_strategy import Inner, Outer
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier import SystemModifier
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination import SystemModifierCombination


class test_SensitivityAnalysisReader(unittest.TestCase):

    def test_generate_system_template_modifier_combination_list(self):
        sensitivity_analysis_dictionary = {
            'SENSITIVITY_ANALYSIS': {
                0: {"PARAMETER_NAME": 'FIRST_OUTER_PARAM_COLUMN/DUMBER',
                    "VALUES": '[4, 5]',
                    "EXPLORATION_STRATEGY": "OUTER"},
                1: {"PARAMETER_NAME": 'SECOND_OUTER_PARAM_COLUMN/DUMBERFailure',
                    "VALUES": '[1., 2.]',
                    "EXPLORATION_STRATEGY": "OUTER"},
                2: {"PARAMETER_NAME": 'FIRST_INNER_PARAM_COLUMN/INNER_PARAM_LINE',
                    "VALUES": '[2., 3.]',
                    "EXPLORATION_STRATEGY": "INNER"},
                3: {"PARAMETER_NAME": 'SECOND_INNER_PARAM_COLUMN/INNER_PARAM_LINE',
                    "VALUES": '[12., 13.]',
                    "EXPLORATION_STRATEGY": "INNER"}
            }
        }

        expected_combination_results = [

            SystemModifierCombination([
                SystemModifier("FIRST_INNER_PARAM_COLUMN/INNER_PARAM_LINE", 2.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 4.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 1.)
            ]),
            SystemModifierCombination([
                SystemModifier("FIRST_INNER_PARAM_COLUMN/INNER_PARAM_LINE", 2.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 4.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 2.)
            ]),
            SystemModifierCombination([
                SystemModifier("FIRST_INNER_PARAM_COLUMN/INNER_PARAM_LINE", 2.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 5.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 1.)
            ]),
            SystemModifierCombination([
                SystemModifier("FIRST_INNER_PARAM_COLUMN/INNER_PARAM_LINE", 2.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 5.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 2.)
            ]),
            SystemModifierCombination([
                SystemModifier("FIRST_INNER_PARAM_COLUMN/INNER_PARAM_LINE", 3.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 4.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 1.)
            ]),
            SystemModifierCombination([
                SystemModifier("FIRST_INNER_PARAM_COLUMN/INNER_PARAM_LINE", 3.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 4.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 2.)
            ]),
            SystemModifierCombination([
                SystemModifier("FIRST_INNER_PARAM_COLUMN/INNER_PARAM_LINE", 3.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 5.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 1.)
            ]),
            SystemModifierCombination([
                SystemModifier("FIRST_INNER_PARAM_COLUMN/INNER_PARAM_LINE", 3.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 5.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 2.)
            ]),



            SystemModifierCombination([
                SystemModifier("SECOND_INNER_PARAM_COLUMN/INNER_PARAM_LINE",12.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 4.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 1.)
            ]),
            SystemModifierCombination([
                SystemModifier("SECOND_INNER_PARAM_COLUMN/INNER_PARAM_LINE", 12.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 4.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 2.)
            ]),
            SystemModifierCombination([
                SystemModifier("SECOND_INNER_PARAM_COLUMN/INNER_PARAM_LINE", 12.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 5.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 1.)
            ]),
            SystemModifierCombination([
                SystemModifier("SECOND_INNER_PARAM_COLUMN/INNER_PARAM_LINE",12.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 5.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 2.)
            ]),
            SystemModifierCombination([
                SystemModifier("SECOND_INNER_PARAM_COLUMN/INNER_PARAM_LINE", 13.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 4.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 1.)
            ]),
            SystemModifierCombination([
                SystemModifier("SECOND_INNER_PARAM_COLUMN/INNER_PARAM_LINE", 13.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 4.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 2.)
            ]),
            SystemModifierCombination([
                SystemModifier("SECOND_INNER_PARAM_COLUMN/INNER_PARAM_LINE", 13.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 5.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 1.)
            ]),
            SystemModifierCombination([
                SystemModifier("SECOND_INNER_PARAM_COLUMN/INNER_PARAM_LINE", 13.),
                SystemModifier("FIRST_OUTER_PARAM_COLUMN/DUMBER", 5.),
                SystemModifier("SECOND_OUTER_PARAM_COLUMN/DUMBERFAILURE", 2.)
            ]),


            ]

        with patch.object(SensitivityAnalysisReader, "read", return_value=sensitivity_analysis_dictionary):
            result = SensitivityAnalysisReader.generate_system_modifier_combination_list("fake_path")

        self.assertEqual(expected_combination_results, result)

    def test_read_exploration_strategy_list(self):
        sensitivity_analysis_dictionary = {
            'SENSITIVITY_ANALYSIS': {
                0: {"PARAMETER_NAME": 'COMPONENT_NUMBER/DUMBER',
                    "VALUES": '[4, 5]',
                    "EXPLORATION_STRATEGY": "OUTER"},
                1: {"PARAMETER_NAME": 'FAILURE_PARAMETERS/DUMBERFAILURE',
                    "VALUES": '[1., 2.]',
                    "EXPLORATION_STRATEGY": "OUTER"},
                2: {"PARAMETER_NAME": 'FAILURE_PARAMETERS/DUMBERFAILURE',
                    "VALUES": '[2., 3.]',
                    "EXPLORATION_STRATEGY": "INNER"}
            }
        }

        expected_value = [
            Outer('COMPONENT_NUMBER/DUMBER', [4, 5]),
            Outer('FAILURE_PARAMETERS/DUMBERFAILURE', [1., 2.]),
            Inner('FAILURE_PARAMETERS/DUMBERFAILURE', [2., 3.]),
        ]

        with patch.object(SensitivityAnalysisReader, "read", return_value=sensitivity_analysis_dictionary):
            result = SensitivityAnalysisReader._read_exploration_strategy_list("fake_path")

        self.assertEqual(expected_value, result)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest
from unittest.mock import patch

from availsim4core.src.context.context import Context
from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.context.system.children_logic.and_ import And
from availsim4core.src.context.system.children_logic.children_logic import ChildrenLogic
from availsim4core.src.context.system.children_logic.oo import Oo
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.component_tree.compound import Compound
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.failure import Failure, FailureType
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.context.system.probability_law.probability_law import ProbabilityLaw
from availsim4core.src.discrete_event_simulation.b_event_generator import BEventGenerator
from availsim4core.src.discrete_event_simulation.c_event_generator import CEventGenerator
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.detectable_failure_event import \
    DetectableFailureEvent
from availsim4core.src.discrete_event_simulation.event.b_event.phase_event.next_phase_event import NextPhaseEvent


class test_Compound(unittest.TestCase):

    def test_to_set(self):
        compound = Compound(1, "test_compound_1", 1, [], ChildrenLogic(), [])

        basic_1 = Basic(2, "test_basic_1", 0, [compound], None, [], [])
        basic_2 = Basic(3, "test_basic_2", 0, [compound], None, [], [])

        component_list = [basic_1, basic_2]
        compound.add_children_list(component_list)

        result = compound.to_set()
        self.assertEqual(result, set([compound, basic_1, basic_2]))

    def test_get_initial_failure_events(self):
        phase_set = set()
        rca_set = set()
        # failures modes used to define a {Basic}
        failure_mode_1 = FailureMode("test_failure_mode_1",
                                     ProbabilityLaw("dummy", 1.0, False),
                                     ProbabilityLaw("dummy", 2.0, False),
                                     Failure(FailureType.DETECTABLE),
                                     phase_set,
                                     None,
                                     phase_set,
                                     None,
                                     'AFTER_REPAIR',
                                     phase_set)
        failure_mode_2 = FailureMode("test_failure_mode_2",
                                     ProbabilityLaw("dummy", 3.0, False),
                                     ProbabilityLaw("dummy", 4.0, False),
                                     Failure(FailureType.DETECTABLE),
                                     phase_set,
                                     None,
                                     phase_set,
                                     None,
                                     'AFTER_REPAIR',
                                     phase_set)

        basic_1 = Basic(1, "test_basic_1", 0, [], failure_mode_1, [], [])
        basic_2 = Basic(2, "test_basic_2", 0, [], failure_mode_2, [], [])

        component_list = [basic_1, basic_2]
        compound = Compound(3, "test_compound_1", 1, [], ChildrenLogic(), [])
        compound.add_children_list(component_list)

        context = Context(compound, PhaseManager(phase_set, set()), rca_set)

        # mocking the get_random_value inside the probability law
        # such mocking is necessary as without it, the random value returned by "get_random_value" can not be predicted
        # the test would always fail
        with patch.object(ProbabilityLaw, "get_random_value", return_value=10.):

            # populate list of B events
            b_events_set = BEventGenerator.generate_first_events(0, context)
            c_events_set = CEventGenerator.generate_first_events(0, context)

            b_events_from_c_events_set = set()
            for c_event in c_events_set:
                if c_event.is_condition_valid():
                    b_events_from_c_events_set.update(c_event.generate_b_events(0))
            b_events_set.update(b_events_from_c_events_set)

        # generating expected events
        import numpy as np
        phase_event = NextPhaseEvent(np.inf, context)
        failure_event_1 = DetectableFailureEvent(10.,
                                                 context,
                                                 basic_1,
                                                 failure_mode_1)
        failure_event_2 = DetectableFailureEvent(10.,
                                                 context,
                                                 basic_2,
                                                 failure_mode_2)

        # testing the equality between result and expected results
        self.assertEqual(b_events_set, set([failure_event_1, failure_event_2, phase_event]))

    def test_get_initial_failure_events__with_absolute_time_not_0(self):
        ABSOLUTE_TIME = 5
        RANDOM_VALUE = 10.
        phase_set = set()
        rca_set = set()
        # failures modes used to define a Basic
        failure_mode_1 = FailureMode("test_failure_mode_1",
                                     ProbabilityLaw("dummy", 1.0, False),
                                     ProbabilityLaw("dummy", 2.0, False),
                                     Failure(FailureType.DETECTABLE),
                                     phase_set,
                                     None,
                                     phase_set,
                                     None,
                                     'AFTER_REPAIR',
                                     phase_set)
        failure_mode_2 = FailureMode("test_failure_mode_2",
                                     ProbabilityLaw("dummy", 3.0, False),
                                     ProbabilityLaw("dummy", 4.0, False),
                                     Failure(FailureType.DETECTABLE),
                                     phase_set,
                                     None,
                                     phase_set,
                                     None,
                                     'AFTER_REPAIR',
                                     phase_set)

        basic_1 = Basic(1, "test_basic_1", 0, [], failure_mode_1, [], [])
        basic_2 = Basic(2, "test_basic_2", 0, [], failure_mode_2, [], [])

        component_list = [basic_1, basic_2]
        compound = Compound(3, "test_compound_1", 1, [], ChildrenLogic(), [])
        compound.add_children_list(component_list)

        context = Context(compound, PhaseManager(phase_set, set()), rca_set)

        # mocking the get_random_value inside the probability law
        # such mocking is necessary as without it, the random value returned by "get_random_value" can not be predicted
        # the test would always fail
        with patch.object(ProbabilityLaw, "get_random_value", return_value=RANDOM_VALUE):

            # populate list of B events
            b_events_set = BEventGenerator.generate_first_events(ABSOLUTE_TIME, context)
            c_events_set = CEventGenerator.generate_first_events(ABSOLUTE_TIME, context)

            b_events_from_c_events_set = set()
            for c_event in c_events_set:
                if c_event.is_condition_valid():
                    b_events_from_c_events_set.update(c_event.generate_b_events(ABSOLUTE_TIME))
            b_events_set.update(b_events_from_c_events_set)

        # generating expected events
        import numpy as np
        phase_event = NextPhaseEvent(np.inf, context)
        failure_event_1 = DetectableFailureEvent(RANDOM_VALUE + ABSOLUTE_TIME,
                                                 context,
                                                 basic_1,
                                                 failure_mode_1)
        failure_event_2 = DetectableFailureEvent(RANDOM_VALUE + ABSOLUTE_TIME,
                                                 context,
                                                 basic_2,
                                                 failure_mode_2)

        # testing the equality between result and expected results
        expected_result = set([failure_event_1, failure_event_2, phase_event])
        self.assertEqual(b_events_set, expected_result)

    def test_update_status__children_running(self):
        compound = Compound(1, "test_update_status", 0, [], And(), [])
        basic_1 = Basic(2, "test_basic_1", 1, [compound], None, [], [])
        basic_2 = Basic(3, "test_basic_2", 2, [compound], None, [], [])
        component_list = [basic_1, basic_2]
        compound.add_children_list(component_list)

        context = Context(None, None, None)

        compound.update_status(10, PhaseManager.NO_PHASE, "test_desc",context)

        self.assertEqual(compound.status, Status.RUNNING)

    def test_update_status__running(self):
        compound = Compound(1, "test_update_status", 0, [], And(), [])
        basic_1 = Basic(2, "test_basic_1", 1, [compound], None, [], [])
        basic_2 = Basic(3, "test_basic_2", 2, [compound], None, [], [])
        component_list = [basic_1, basic_2]
        compound.add_children_list(component_list)
        basic_2.status = Status.FAILED

        context = Context(None,None,None)

        compound.update_status(10, PhaseManager.NO_PHASE, "test_desc",context)

        self.assertEqual(compound.status, Status.FAILED)

    def test_evaluate_status_and_running(self):
        compound = Compound(1, "test_compound_1", 0, [], And(), [])

        basic_1 = Basic(2, "test_basic_1", 1, [compound], None, [], [])
        basic_2 = Basic(3, "test_basic_2", 2, [compound], None, [], [])

        component_list = [basic_1, basic_2]
        compound.add_children_list(component_list)

        context = Context(None,None,None)

        result = compound.evaluate_status(context)

        self.assertEqual(result, Status.RUNNING)

    def test_evaluate_status_and_failed(self):
        compound = Compound(1, "test_compound_1", 0, [], And(), [])

        basic_1 = Basic(2, "test_basic_1", 1, [compound], None, [], [])
        basic_2 = Basic(3, "test_basic_2", 2, [compound], None, [], [])
        basic_1.status = Status.FAILED
        basic_2.status = Status.FAILED

        component_list = [basic_1, basic_2]
        compound.add_children_list(component_list)

        context = Context(None,None,None)

        result = compound.evaluate_status(context)

        self.assertEqual(result, Status.FAILED)

    def test_evaluate_status_1oo2_running(self):
        compound = Compound(1, "test_compound_1", 0, [], Oo(1,2), [])

        basic_1 = Basic(2, "test_basic_1", 1, [compound], None, [], [])
        basic_2 = Basic(3, "test_basic_2", 2, [compound], None, [], [])
        basic_1.status = Status.FAILED
        basic_2.status = Status.RUNNING

        component_list = [basic_1, basic_2]
        compound.add_children_list(component_list)

        context = Context(None,None,None)

        result = compound.evaluate_status(context)

        self.assertEqual(result, Status.DEGRADED)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import numpy
from numpy.random import binomial

from availsim4core.src.context.system.probability_law.probability_law import ProbabilityLaw


class BinomialLaw(ProbabilityLaw):
    """
    Class used to generate values according to the Binomial law
    """

    def __init__(self,
                 parameters):
        super().__init__(self.__class__.__name__,
                         parameters,
                         is_failure_on_demand=True)

    def __eq__(self, other):
        if not isinstance(other, BinomialLaw):
            return NotImplemented
        return super().__eq__(other)

    def get_random_value(self):
        return binomial(1, self.parameters[0])

    def get_quantile_value(self, quantile):
        return quantile <= self.parameters[0]

    def get_mean_value(self):
        return self.parameters[0]

    def get_variance_value(self):
        return self.parameters[0]*(1-self.parameters[0])



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.results.result_record_entry_phase import ResultRecordEntryPhase
from availsim4core.src.timeline.record import Record


class RecordPhase(Record):
    """
    Class of record dedicated to phases
    A record is an object used to keep trace of the state of the system
    """

    def __init__(self,
                 phase: Phase,
                 timestamp: float,
                 description: str):
        super().__init__(timestamp, phase, description)

    def __eq__(self, other):
        if not isinstance(other, RecordPhase):
            return NotImplemented
        return super().__eq__(other)

    def __repr__(self):
        return super().__repr__() + f" phase: {self.phase}"

    def __str__(self):
        return f"RecordPhase at {self.timestamp}: {self.phase.name}"

    def get_result_record_entry(self):
        return ResultRecordEntryPhase(self.phase, self.description)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.discrete_event_simulation.event.b_event.inspection_event.start_inspection_event import \
    StartInspectionEvent


class InspectionEventFactory:

    @staticmethod
    def build(absolute_occurrence_time: float,
              context: Context,
              basic: Basic,
              failure_mode: FailureMode):
        """
        Build a StartInspectionEvent.
        :param absolute_occurrence_time: absolute discrete event simulation time when the inspection event will be started.
        :param context:
        :param basic: associate basic to inspect.
        :param failure_mode: failure mode which triggered this inspection event.
        """
        if failure_mode.inspection is None:
            return None
        return StartInspectionEvent(absolute_occurrence_time + failure_mode.inspection.periodicity,
                                    context,
                                    basic,
                                    failure_mode)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import Set

import numpy

from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.system.phase_jump_trigger import PhaseJumpTrigger
from availsim4core.src.context.system.probability_law.deterministic_law import DeterministicLaw
from availsim4core.src.timeline.record_phase import RecordPhase


class PhaseManager:
    """
    This class provides all the methods to manipulate {Phase}.
    Note : When no {Phase} is defined in the simulation. The NO_PHASE is used by default.
    """
    __slots__ = 'phase_set', 'current_phase', 'phase_jump_trigger_set'

    NO_PHASE = Phase("NONE", DeterministicLaw([numpy.inf]), False)
    NO_PHASE.set_next_phase(NO_PHASE)
    NO_PHASE.set_next_phase_if_failure(NO_PHASE)

    HELD_FOREVER = Phase("HELD_FOREVER", DeterministicLaw([numpy.inf]), False)
    HELD_FOREVER.set_next_phase(HELD_FOREVER)
    HELD_FOREVER.set_next_phase_if_failure(HELD_FOREVER)

    DEFAULT_PHASE_DESCRIPTION = "default"
    DEFAULT_PHASE_IF_FAILURE_DESCRIPTION = "default failure"
    SPECIFIC_PHASE_IF_FAILURE_DESCRIPTION = "specific failure"
    JUMP_INTO_PHASE_DESCRIPTION = "jump into phase"
    UPDATE_FOLLOWING_A_PHASE_CHANGE = "Update following a phase change"

    def __init__(self,
                 phase_set: Set[Phase],
                 phase_jump_trigger_set: Set[PhaseJumpTrigger]):
        self.phase_set: Set[Phase] = phase_set
        self.current_phase = self._get_first_phase()
        self.phase_jump_trigger_set = phase_jump_trigger_set

    def __eq__(self, other):
        return self.phase_set == other.phase_set and \
               self.current_phase == other.current_phase and \
               self.phase_jump_trigger_set == other.phase_jump_trigger_set


    def __str__(self):
        return f"Current Phase -> {self.current_phase}"

    def _get_first_phase(self) -> Phase:
        """
        Find the first {Phase} of the simulation.
        :return: The first {Phase} of the simulation, NO_PHASE otherwise.
        """
        return next((phase
                     for phase in self.phase_set
                     if phase.is_first_phase), PhaseManager.NO_PHASE)

    def _find_phase(self, phase_name_to_find) -> Phase:
        """
        Given a string name of phase, this methods returns the corresponding {Phase} is defined in the init list.
        If the `phase_name_to_find` is not found, returns the NO_PHASE.
        :param phase_name_to_find: the str name of the phase to find.
        :return: {Phase} with the given `phase_name_to_find`, NO_PHASE otherwise.
        """
        return next((phase
                     for phase in self.phase_set
                     if phase.name == phase_name_to_find), PhaseManager.NO_PHASE)

    def go_to_phase(self,
                    phase_start_time,
                    phase: Phase,
                    description: str) -> RecordPhase:
        """
        Updating the "current_phase" of the phase manager
        """
        self.current_phase = phase
        return RecordPhase(self.current_phase,
                           phase_start_time,
                           description)


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import TYPE_CHECKING, Set, Tuple, List

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.basic import Basic

from availsim4core.src.context.context import Context
from availsim4core.src.discrete_event_simulation.event.b_event.b_event import BEvent, BEventPriority


class BasicBEvent(BEvent):
    """
    Class dealing with B events related to basic components
    """
    __slots__ = 'basic'

    def __init__(self,
                 absolute_occurrence_time: float,
                 context: Context,
                 basic: Basic,
                 priority: BEventPriority):
        super().__init__(absolute_occurrence_time, context, priority)
        self.basic = basic

    def __eq__(self, other):
        return super().__eq__(other) and \
               self.basic == other.basic

    def __lt__(self, other):
        if not isinstance(other, BasicBEvent):
            return super().__lt__(other)
        else:
            return super().__lt__(other) or (super().__eq__(other) and self.basic.global_id < other.basic.global_id)


    def update_b_event_collection(self,
                                event_set: Set,
                                types_of_event_to_clean: List[type]) -> Tuple[Set, Set]:
        """
        The input event_set is cleaned from all the events of types provided by types_of_event_to_clean.
        Returns a clean set of Events and a set containing the removed events.
        """
        if not types_of_event_to_clean:
            return event_set, set()

        event_to_remove = set()

        for event in event_set:
            if type(event) in types_of_event_to_clean:

                # if the event has a basic
                if hasattr(event,'basic'):
                    # and the basic is similar to the self.basic
                    if event.basic == self.basic:
                        # then this event has to be removed
                        event_to_remove.add(event)
                else:
                # if the event has no basic, then it has to be removed
                    event_to_remove.add(event)

        return event_set - event_to_remove, event_to_remove



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import List

from availsim4core.src.timeline.record import Record
from availsim4core.src.timeline.record_component import RecordComponent


class Timeline:
    """
    Timeline of records which have been proceed in the simulation.
    The timeline is ordered.
    """

    def __init__(self):
        self.record_list: List[Record] = []

    def __eq__(self,other):
        return self.record_list == other.record_list

    def __str__(self):
        return str(self.record_list)

    def __repr__(self):
        return self.__str__()

    def extract_record_from_types(self, types):
        """
        Given a tuple a class types this method returns corresponding list of record matching this types.
        :param types: Class types to be filtered out from the list.
        :return: the list of records corresponding to the given types.
        """
        return [record
                for record in self.record_list
                if isinstance(record, types)]

    def _get_previous_record_of_type(self, specified_type):
        """
        Function getting the previous record of a given type.
        :param specified_type: type of the desired record
        :return: None if no record have been found, otherwise the record founded
        """
        counter = 0
        for record in reversed(self.record_list):
            if isinstance(record, specified_type):
                counter += 1
                if counter == 2:
                    return record
        return None

    def _get_previous_record_of_basic_in_status(self, basic, status):
        """
        Function getting the previous record linked to a specific basic component in a specific status.
        :param basic: component to find back in the list of event
        :param status: status to find back in the list of event
        :return: None if no record have been found, otherwise the record founded
        """
        counter = 0
        for record in reversed(self.record_list):
            if isinstance(record,RecordComponent):
                if record.component == basic:
                    # the list of events in the inputs all have a 'basic' attribute because they have failures event
                    # attached to them that one wants to postpone
                    counter += 1
                    if (counter >= 2) and (record.status == status):
                        return record
        return None


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 


import copy
import logging
from datetime import datetime
from typing import List

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.discrete_event_simulation.discrete_event_simulation import DiscreteEventSimulation
from availsim4core.src.results.simulation_results import SimulationResults
from availsim4core.src.simulation.monte_carlo import MonteCarlo
from availsim4core.src.simulation.simulation import Simulation, SimulationType
from availsim4core.src.statistics.critical_failure_paths import CriticalFailurePaths


class SplittingMonteCarlo(MonteCarlo):
    """
    Defines a MonteCarlo simulation that will be triggered by the Sensitivity Analysis.
    Attributes:
    - minimum_number_of_simulations: minimun number of call to discrete event simulation.
    - maximum_number_of_simulations: maximun number of call to discrete event simulation.
    - convergence_margin: Evaluate on the fly the appropriate number of simulation to run.
    - maximum_execution_time: used on the Cluster to stop the simulation after this period of time.
    - seed: Control the random generator to get different values over each simulation. Deterministic approach.
    - list_of_diagnosis: Defines the metrics to compute.
    - result_simulation: The ResultSimulation to return.
    Extends the simulation, by implementing the run() method.
    """
    SIMULATION_TYPE = SimulationType.SPLITTING_MONTE_CARLO

    def __init__(self, minimum_number_of_simulations: int, 
                 maximum_number_of_simulations: int, 
                 convergence_margin: float, 
                 maximum_execution_time: float, 
                 seed: int, 
                 list_of_diagnosis: List[str], 
                 duration_of_simulation: float, 
                 system_root_component_name: str):
        self.system_root_component_name = system_root_component_name
        super().__init__(minimum_number_of_simulations, 
                         maximum_number_of_simulations, 
                         convergence_margin, 
                         maximum_execution_time, 
                         seed, 
                         list_of_diagnosis, 
                         duration_of_simulation)

    def run(self, context: Context):
        """
        This method is running multiple iterations of the model and updating statistics according to the results. The function
        provides means to repeat partially completed timelines (when DES function returns a timeline which has not reached the
        user-defined simulation duration value). Since in the splitting method the number of model evaluations is not known and 
        changes dynamically, the method is based on a queue which gathers information about consecutive repeated timelines. 
        """
        simulation_results = SimulationResults(self.maximum_number_of_simulations,
                                               self.duration_of_simulation)

        # Those two parameters are for now hardcoded below. In the long run, they are to be optionally 
        # defined by users or determined in an automatic way (adaptive execution, pilot jobs, etc.). 
        restarts_in_level = [10, 10, 10, 10]
        threshold = 1

        # Critical failure paths are used to establish the distance to a critical failure inside the DES
        cfp = CriticalFailurePaths(context, self.system_root_component_name)

        start_time = datetime.now()
        logging.info(f"Starting a SPLITTING Monte Carlo simulation {0+1} / {self.maximum_number_of_simulations}")

        # Initialization of the queue with the parameters creating the first level (equivalent to standard MC)
        queue = [{"context": context, 
                  "simulation_absolute_time": 0, 
                  "simulation_weight": 1/self.maximum_number_of_simulations, 
                  "simulation_level_threshold": threshold, 
                  "simulations_to_execute": self.maximum_number_of_simulations,
                  "recursion_depth": 0}]

        evaluations_counter = 0
        while queue:
            # unpack the parameters for a given queue element
            parameters = queue[0]
            context = parameters["context"]
            simulation_absolute_time = parameters["simulation_absolute_time"]
            simulation_weight = parameters["simulation_weight"]
            level_threshold = parameters["simulation_level_threshold"]
            recursion_depth = parameters["recursion_depth"]

            # initialisation of a DES
            simulation_des = DiscreteEventSimulation(self.seed + evaluations_counter,
                                                     self.duration_of_simulation,
                                                     copy.deepcopy(context),
                                                     cfp,
                                                     level_threshold,
                                                     simulation_absolute_time)

            # simulation
            context, simulation_absolute_time = simulation_des.run()

            if simulation_absolute_time < self.duration_of_simulation:
                # restarting
                queue.append({"context": copy.deepcopy(context), 
                    "simulation_absolute_time": simulation_absolute_time, 
                    "simulation_weight": simulation_weight / restarts_in_level[recursion_depth+1], 
                    "simulation_level_threshold": level_threshold + 1, 
                    "simulations_to_execute": restarts_in_level[recursion_depth+1],
                    "recursion_depth": recursion_depth+1})
            else:
                timeline_record = simulation_des.context.timeline_record

                # updating statistics
                self.update_statistics(simulation_results,
                                       simulation_des,
                                       start_time,
                                       simulation_weight,
                                       recursion_depth)

            evaluations_counter += 1
            parameters["simulations_to_execute"] -= 1
            if parameters["simulations_to_execute"] == 0:
                del queue[0]

        simulation_results.evaluate_result()
        simulation_results.last_simulation_timeline = timeline_record.record_list
        return simulation_results



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.children_logic.children_logic import ChildrenLogic
from availsim4core.src.context.system.children_logic.children_logic_factory import ChildrenLogicFactory
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.component_tree.compound import Compound
from availsim4core.src.context.system.component_tree.status import Status


# TODO: add tests on the status UNDER_REPAIR, BLIND_FAILED, etc

class test_TF(unittest.TestCase):

    context = Context(None, None, None)

    # number_of_running_basic, number_of_failed_basic, number_of_running_compound, number_of_failed_compound
    scenario_1 = [8, 0, 0, 0, "2TF", context, Status.RUNNING]
    scenario_2 = [6, 2, 0, 0, "2TF", context, Status.DEGRADED]
    scenario_3 = [4, 4, 0, 0, "2TF", context, Status.FAILED]
    scenario_4 = [0, 0, 8, 0, "2TF", context, Status.RUNNING]
    scenario_5 = [0, 0, 6, 2, "2TF", context, Status.DEGRADED]
    scenario_6 = [0, 0, 4, 4, "2TF", context, Status.FAILED]
    scenario_7 = [4, 0, 4, 0, "2TF", context, Status.RUNNING]
    scenario_8 = [3, 1, 4, 0, "2TF", context, Status.DEGRADED]
    scenario_9 = [2, 2, 2, 2, "2TF", context, Status.FAILED]
    param_scenario = [scenario_1, scenario_2, scenario_3,
                      scenario_4, scenario_5, scenario_6,
                      scenario_7, scenario_8, scenario_9]

    def _evaluate_generic(self,
                          number_of_running_basic: int,
                          number_of_failed_basic: int,
                          number_of_running_compound: int,
                          number_of_failed_compound: int,
                          children_logic_str: str,
                          context: Context,
                          expected_result: Status):

        list_of_component = []

        for i in range(number_of_running_basic):
            basic = Basic(0, "dummyName", 0, [], [], [], [])
            list_of_component.append(basic)

        for i in range(number_of_failed_basic):
            basic = Basic(0, "dummyName", 0, [], [], [], [])
            basic.status = Status.FAILED
            list_of_component.append(basic)

        for i in range(number_of_running_compound):
            compound = Compound(0, "test_compound_1", 1, [], ChildrenLogic(), [])
            list_of_component.append(compound)

        for i in range(number_of_failed_compound):
            compound = Compound(0, "test_compound_1", 1, [], ChildrenLogic(), [])
            compound.status = Status.FAILED
            list_of_component.append(compound)

        logic = ChildrenLogicFactory.build(children_logic_str, None)
        result = logic.evaluate(list_of_component, context)

        self.assertEqual(result, expected_result)

    def test_runner(self):
        for number_of_running_basic, number_of_failed_basic, \
            number_of_running_compound, number_of_failed_compound, \
            logic, context, expected_result in self.param_scenario:
            with self.subTest():
                self._evaluate_generic(number_of_running_basic,
                                       number_of_failed_basic,
                                       number_of_running_compound,
                                       number_of_failed_compound,
                                       logic,
                                       context,
                                       expected_result)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import List, Set

from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.system.architecture_entry import ArchitectureEntry
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.context.system.failure_mode_assignments import FailureModeAssignments
from availsim4core.src.context.system.inspection import Inspection
from availsim4core.src.context.system.minimal_replaceable_unit import MinimalReplaceableUnit
from availsim4core.src.context.system.phase_jump_trigger import PhaseJumpTrigger
from availsim4core.src.context.system.rca_trigger import RootCauseAnalysisTrigger


class SystemTemplate:
    """
    Domain object to represent the configuration template entries provided by the users.
    """
    __slots__ = 'architecture_entry_list', \
                'failure_mode_assignments_list', \
                'failure_mode_list', \
                'mru_list', \
                'inspection_list', \
                'phase_set', \
                'root_cause_analysis_trigger_set', \
                'phase_jump_trigger_set', \
                'custom_children_logic_path'

    def __init__(self,
                 architecture_entry_list: List[ArchitectureEntry],
                 failure_mode_assignments_list: List[FailureModeAssignments],
                 failure_mode_list: List[FailureMode],
                 mru_list: List[MinimalReplaceableUnit],
                 inspection_list: List[Inspection],
                 phase_set: Set[Phase],
                 root_cause_analysis_trigger_set: Set[RootCauseAnalysisTrigger],
                 phase_jump_trigger_set: Set[PhaseJumpTrigger],
                 custom_children_logic_path: str):
        self.architecture_entry_list = architecture_entry_list
        self.failure_mode_assignments_list = failure_mode_assignments_list
        self.failure_mode_list = failure_mode_list
        self.mru_list = mru_list
        self.inspection_list = inspection_list
        self.phase_set = phase_set
        self.root_cause_analysis_trigger_set = root_cause_analysis_trigger_set
        self.phase_jump_trigger_set = phase_jump_trigger_set
        self.custom_children_logic_path = custom_children_logic_path

    def __eq__(self, other):
        return self.architecture_entry_list == other.architecture_entry_list \
               and self.failure_mode_assignments_list == other.failure_mode_assignments_list \
               and self.failure_mode_list == other.failure_mode_list \
               and self.mru_list == other.mru_list \
               and self.inspection_list == other.inspection_list \
               and self.phase_set == other.phase_set \
               and self.root_cause_analysis_trigger_set == other.root_cause_analysis_trigger_set \
               and self.phase_jump_trigger_set == other.phase_jump_trigger_set \
               and self.custom_children_logic_path == other.custom_children_logic_path

    def find_architecture_entry(self, name):
        return next((entry for entry in self.architecture_entry_list if entry.component_name == name), None)

    def find_failure_mode(self, name):
        return next((failure_mode for failure_mode in self.failure_mode_list if failure_mode.name == name), None)

    def find_mru(self, name):
        return next((mru for mru in self.mru_list if mru.name == name), None)

    def find_inspection(self, name):
        return next((inspection for inspection in self.inspection_list if inspection.name == name), None)

    def find_phase(self, name):
        return next((phase for phase in self.phase_set if phase.name == name), None)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.system.failure_mode import FailureMode


class FailureModeAssignments:
    __slots__ = 'component_name', 'failure_mode'

    def __init__(self,
                 component_name: str,
                 failure_mode: FailureMode):
        self.component_name = component_name
        self.failure_mode = failure_mode

    def __eq__(self, other):
        if not isinstance(other, FailureModeAssignments):
            return NotImplemented
        return self.component_name == other.component_name and \
               self.failure_mode == other.failure_mode

    def __repr__(self):
        return f"component_name: {self.component_name} -> " \
               f"failure_mode.name: {self.failure_mode.name}"



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.sensitivity_analysis.exploration_strategy import Inner, Outer
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier import SystemModifier
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination import SystemModifierCombination
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination_strategy_director import \
    SystemModifierCombinationStrategyDirector


class test_SystemModifierCombinationStrategyDirector(unittest.TestCase):

    def test_execute__outer_inner(self):
        system_modifier_list_outer_1 = [SystemModifier("inner", 5.), SystemModifier("outer1", 1.), SystemModifier("outer2", 3.)]
        system_modifier_list_outer_2 = [SystemModifier("inner", 5.), SystemModifier("outer1", 1.), SystemModifier("outer2", 4.)]
        system_modifier_list_outer_3 = [SystemModifier("inner", 5.), SystemModifier("outer1", 2.), SystemModifier("outer2", 3.)]
        system_modifier_list_outer_4 = [SystemModifier("inner", 5.), SystemModifier("outer1", 2.), SystemModifier("outer2", 4.)]

        expected_combination_result = [SystemModifierCombination(system_modifier_list_outer_1),
                                       SystemModifierCombination(system_modifier_list_outer_2),
                                       SystemModifierCombination(system_modifier_list_outer_3),
                                       SystemModifierCombination(system_modifier_list_outer_4)]

        exploration_strategy_list = [Outer("outer1", [1., 2.]),
                                     Outer("outer2", [3., 4.]),
                                     Inner("inner", [5.])]

        result = SystemModifierCombinationStrategyDirector.execute(exploration_strategy_list)

        self.assertEqual(expected_combination_result, result)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

import argparse
import logging
import logging.config
import os
import pathlib

from availsim4core._version import __version__

LOGGING_CONFIG_FILE = '/logging/logging.conf'


def logger_config(debug: bool, output_folder: str) -> str:
    """
    Loads the configuration from the `LOGGING_CONFIG_FILE`.
    :param debug: If set the root logger level is set to debug
    :param output_folder: the output folder where the log will be written.
    :return: The path of the first `FileHandler` where the logs are written.
    """
    # using logging namespace
    logging.output_folder = output_folder
    
    path = pathlib.Path(__file__).parent.absolute()
    logging.config.fileConfig(f"{path}{LOGGING_CONFIG_FILE}")
    if debug:
        logging.getLogger().setLevel(logging.DEBUG)
    return next((handler.baseFilename
                 for handler in logging.getLogger().handlers
                 if isinstance(handler, logging.FileHandler)), None)


def parser_config():
    # read arguments
    parser = argparse.ArgumentParser(description=f"running AvailSim4 -")

    parser.add_argument('--version', action='version',
                        version='%(prog)s {version}'.format(version=__version__))

    parser.add_argument('--simulation', type=str, required=True,
                        help='parameters of the algorithm')

    parser.add_argument('--system', type=str, required=True,
                        help='system file to simulate')

    parser.add_argument('--sensitivity_analysis', type=str, required=False,
                        help='sensitivity analysis file to explore',
                        default=None)

    parser.add_argument('--children_logic', type=str, required=False,
                        help='python file defining a custom children logic',
                        default=None)

    parser.add_argument('--HTCondor', default=False, action='store_true',
                        help='if the flag is present, then HTCondor is used to '
                             'run the different jobs of the sensitivity analysis')

    parser.add_argument('--HTCondor_extra_argument', type=str, required=False, default="",
                        help='possible extra argument used when submitting jobs '
                             'most likely a group/quota definition such as: +AccountingGroup="group_u_XX.XXX"')

    parser.add_argument('--nb_processes', type=int, required=False,
                        help='number of processes used in the parallel part of the code '
                             '(over the sensitivity analysis when HTCondor is not used)',
                        default=1)

    parser.add_argument('--output_folder', type=str, required=True,
                        help='folder in which the results are exported')

    parser.add_argument('--debug', default=False, action='store_true',
                        help='if the flag is present, then debug mode is used in the logging file')

    return parser.parse_args()


def init():
    """
    Parse the command arguments and loads the logging configuration.
    :return tuple coming from given argument
    simulation_path, system_path, sensitivity_analysis_path, output_folder, HTCondor, nb_process
    """
    args = parser_config()

    os.makedirs(args.output_folder, exist_ok=True)

    log_file_path = logger_config(args.debug, args.output_folder)

    logging.info(f"- Availsim4 - version: {__version__}")
    logging.info(f"- Init configuration -")

    if args.debug:
        if log_file_path is None:
            logging.error("DEBUG mode activated but logging file not defined or misconfigured.")
        logging.info(f"DEBUG mode activated -> Check: {log_file_path}")

    logging.debug(f"Arguments: "
                  f"simulation:{args.simulation} - "
                  f"system:{args.system} - "
                  f"sensitivity analysis:{args.sensitivity_analysis} - "
                  f"output_folder:{args.output_folder} - "
                  f"HTCondor:{args.HTCondor} - "
                  f"HTCondor_extra_argument:{args.HTCondor_extra_argument} - "
                  f"nb_processes:{args.nb_processes} - "
                  f"children_logic:{args.children_logic}")

    return (args.simulation,
            args.system,
            args.sensitivity_analysis,
            args.output_folder,
            args.HTCondor,
            args.HTCondor_extra_argument,
            args.nb_processes,
            args.children_logic)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import Set, TYPE_CHECKING, Tuple, List

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.basic import Basic

from availsim4core.src.discrete_event_simulation.event.b_event.b_event import BEventPriority
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEvent, CEventPriority
from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.failure_event import FailureEvent
from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.discrete_event_simulation.event.c_event.held_event.order_end_holding_event import \
    OrderEndHoldingEvent
from availsim4core.src.discrete_event_simulation.event.c_event.repair_event.order_repair_event import OrderRepairEvent


class DetectableFailureEvent(FailureEvent):
    """
    Class dealing with detectable failure events. A detectable failure can trigger a repair because that failure is ,
    in real life, detected by the monitoring system
    """

    def __init__(self,
                 absolute_occurrence_time: float,
                 context: Context,
                 basic: Basic,
                 failure_mode: FailureMode):
        super().__init__(absolute_occurrence_time, context, basic, failure_mode,
                         priority=BEventPriority.DETECTABLE_FAILURE_EVENT)

    def postpone(self, duration):
        return DetectableFailureEvent(self.absolute_occurrence_time + duration,
                                      self.context,
                                      self.basic,
                                      self.failure_mode)

    def execute(self):
        return self.basic.update_status(Status.FAILED,
                                        self.absolute_occurrence_time,
                                        self.context.phase_manager.current_phase,  # or phase of the failure mode?!
                                        f"{self.failure_mode.name} failure mode of component {self.basic.name}_{self.basic.local_id}_{self.basic.global_id}",
                                        self.context)

    def update_b_event_collection(self,
                                event_set: Set,
                                types_of_event_to_clean: List[type]) -> Tuple[Set, Set]:
        """
        The input event_set is cleaned from all the events of types provided by types_of_event_to_clean.
        Returns a clean set of Events and a set containing the removed events.
        """

        if self.failure_mode.phase_change_trigger in ["AFTER_REPAIR", "AFTER_FAILURE"]:
            # we clean the flow of the phases as for failures, the tye_of_event_to_clean are only the phase event
            b_event_to_remove_set = {event
                                     for event in event_set
                                     if type(event) in types_of_event_to_clean}
            return event_set - b_event_to_remove_set, b_event_to_remove_set

        elif self.failure_mode.phase_change_trigger in ["NEVER"]:
            # we do not change the normal flow of the phases
            return event_set, set()

    def generate_c_event(self, **kwargs) -> Set[CEvent]:

        c_event_set = set()

        order_repair_event = OrderRepairEvent(
            priority=CEventPriority.ORDER_REPAIR_EVENT,
            context=self.context,
            component=self.basic,
            event=self,
            failure_mode=self.failure_mode
        )

        if self.failure_mode.held_before_repair_phase_set != {PhaseManager.HELD_FOREVER}:
            if self.context.phase_manager.current_phase not in self.failure_mode.held_before_repair_phase_set:
                c_event_set.add(OrderEndHoldingEvent(
                    priority=CEventPriority.ORDER_END_HOLDING_EVENT,
                    context=self.context,
                    component=self.basic,
                    event=self,
                    failure_mode=self.failure_mode,
                    held_event=order_repair_event,
                    held_until_phase_set=self.failure_mode.held_before_repair_phase_set
                ))
            else:
                c_event_set.add(order_repair_event)

        if self.failure_mode.phase_change_trigger in ["AFTER_FAILURE"]:

            from availsim4core.src.discrete_event_simulation.event.c_event.phase_event.order_next_phase_if_failure_event import \
                OrderNextPhaseIfFailureEvent
            c_event_set.add(
                OrderNextPhaseIfFailureEvent(
                    priority=CEventPriority.ORDER_NEXT_PHASE_EVENT,
                    context=self.context,
                    failure_mode=self.failure_mode
                )
            )

        return c_event_set


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import copy
import unittest

from availsim4core.src.context.context import Context
from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.system.children_logic.and_ import And
from availsim4core.src.context.system.component_tree.basic import Basic
from availsim4core.src.context.system.component_tree.compound import Compound
from availsim4core.src.context.system.failure import Failure, FailureType
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.context.system.inspection import Inspection
from availsim4core.src.context.system.probability_law.probability_law import ProbabilityLaw
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.detectable_failure_event import \
    DetectableFailureEvent
from availsim4core.src.discrete_event_simulation.event.b_event.inspection_event.start_inspection_event import \
    StartInspectionEvent


class test_Event(unittest.TestCase):

    def test__clean_event_collection(self):
        context = Context(None, [], set())
        root_component = Compound(1, "ROOT", 0, [], And(), [])
        _basic = Basic(2, "BASIC", 0, [root_component], [], [], [])
        _basic2 = Basic(3, "BASIC", 0, [root_component], [], [], [])

        NO_PHASE = Phase("NONE", ProbabilityLaw("", 0, False), False)
        NO_PHASE.set_next_phase(NO_PHASE)
        NO_PHASE.set_next_phase_if_failure(NO_PHASE)

        expected_inspection = Inspection("inspection_test", 1, 2)

        failure_mode = FailureMode("TEST_FAILURE",
                                   ProbabilityLaw("", 0, False),
                                   ProbabilityLaw("", 0, False),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   expected_inspection,
                                   [],
                                   NO_PHASE,
                                   'AFTER_FAILURE',
                                   []
                                   )

        event_basic_1 = StartInspectionEvent(0, context, _basic, failure_mode)
        event_basic_2 = DetectableFailureEvent(1, context, _basic, copy.deepcopy(failure_mode))
        event_basic_3 = DetectableFailureEvent(1, context, _basic2, copy.deepcopy(failure_mode))

        event_set = {event_basic_1, event_basic_2, event_basic_3}

        result, event_removed_result = event_basic_1.update_b_event_collection(
            event_set,
            event_basic_1._b_events_to_be_cleaned())

        expected_result = {event_basic_1, event_basic_3}
        expected_removed = {event_basic_2}
        self.assertEqual(result, expected_result)
        self.assertEqual(event_removed_result, expected_removed)

    def test__clean_event_collection_not_changed(self):
        context = Context(None, [], set())
        root_component = Compound(1, "ROOT", 0, [], And(), [])
        _basic = Basic(2, "BASIC", 0, [root_component], [], [], [])
        _basic2 = Basic(3, "BASIC", 0, [root_component], [], [], [])

        NO_PHASE = Phase("NONE", ProbabilityLaw("", 0, False), False)
        NO_PHASE.set_next_phase(NO_PHASE)
        NO_PHASE.set_next_phase_if_failure(NO_PHASE)

        expected_inspection = Inspection("inspection_test", 1, 2)

        failure_mode = FailureMode("TEST_FAILURE",
                                   ProbabilityLaw("", 0, False),
                                   ProbabilityLaw("", 0, False),
                                   Failure(FailureType.BLIND),
                                   "[]",
                                   expected_inspection,
                                   [],
                                   NO_PHASE,
                                   'AFTER_FAILURE',
                                   []
                                   )

        event_basic_1 = StartInspectionEvent(0, context, _basic, failure_mode)
        event_basic_2 = DetectableFailureEvent(0, context, _basic2, copy.deepcopy(failure_mode))

        event_set = {event_basic_1, event_basic_2}

        result, event_removed_result = event_basic_1.update_b_event_collection(
            event_set,
            event_basic_1._b_events_to_be_cleaned())

        expected_result = event_set
        self.assertEqual(result, expected_result)
        self.assertEqual(event_removed_result, set())

    # TODO with the mru, inspection, blind, etc ... all kind of events Bevents and Cevents.



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.system.component_tree.component import Component


class SystemUtils:
    # TODO: use regexp instead of manipulating the string ourselves

    @staticmethod
    def is_string_containing_parenthesis(input: str):
        return "(" in input and ")" in input

    @staticmethod
    def extract_arguments_within_parenthesis(input: str):
        """
        format of the string: X(Y)
        X = name of function
        Y = list of arguments
        """
        return input.split("(")[1].split(")")[0]

    @staticmethod
    def extract_name_of_function_from_string(str_input: str):
        """
        format of the string: X(Y)
        X = name of function
        Y = list of arguments
        """
        return str_input.split("(")[0]

    @classmethod
    def find_first_lowest_level_ancestor_with_name(cls,
                                                   component: Component,
                                                   ancestor_name: str):
        """
        Find the lowest level first ancestor parent from the given the component parents and based on the given ancestor name.
        None otherwise.
        :param component: The component to inspect.
        :param ancestor_name: the name of the lowest common parent.
        :return: the ancestor parent of the given component with the name, None otherwise.
        """

        found_ancestor_tuple_list = cls._find_ancestors_and_treelevel_with_name(component,
                                                                                ancestor_name,
                                                                                0)
        return None if not found_ancestor_tuple_list \
            else sorted(found_ancestor_tuple_list, key=lambda ancestor: ancestor[1])[0][0]

    @classmethod
    def _find_ancestors_and_treelevel_with_name(cls,
                                                component: Component,
                                                ancestor_name: str,
                                                tree_level):
        if component.name == ancestor_name:
            return [(component, tree_level)]
        found_ancestor_tuple_list = []
        tree_level += 1
        for parent in component.get_parents():
            found_ancestors = cls._find_ancestors_and_treelevel_with_name(parent,
                                                                          ancestor_name,
                                                                          tree_level)
            found_ancestor_tuple_list.extend(found_ancestors)
        return found_ancestor_tuple_list



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.context import Context
from availsim4core.src.discrete_event_simulation.event.event import Event
from enum import Enum


class CEventPriority(Enum):
    """
    class used to easily order the C Event set
    """

    ORDER_START_INSPECTION_EVENT = -7
    ORDER_END_INSPECTION_EVENT = -6

    ORDER_JUMP_PHASE_EVENT = -5
    ORDER_NEXT_PHASE_EVENT_IF_SPECIFIC_FAILURE = -4
    ORDER_NEXT_PHASE_EVENT = -3

    ORDER_POSTPONE_C_EVENT = -2

    MRU_ORDER_REPAIR_EVENT = -1

    ORDER_FAILURE_EVENT = 0
    REEVALUATE_ORDER_ALL_FAILURE_EVENTS = 1
    REEVALUATE_ORDER_ALL_INSPECTION_EVENTS = 2
    ORDER_REPAIR_EVENT = 3
    ORDER_END_HOLDING_EVENT = 4

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


class CEvent(Event):
    """
    C events = Conditioned events.
    Those events are not planned in advance but only occur when some particular conditions are present. This type of events would be planning of actions in our case, that
    is to say creating B events:  ordering a repair now or at a later date, advancing the end of a phase or the next maintenance period, the next periodic inspection.
    The logic could be complex. Maybe some C events are postponed because conditions are not realised to trigger a B event associated: lack of spare or man power to repair something, ...
    """
    __slots__ = 'priority'

    def __init__(self,
                 priority: CEventPriority,
                 context: Context):
        super().__init__(context)
        self.priority = priority

    def __eq__(self, other):
        return self.priority == other.priority

    def __lt__(self, other):
        return self.priority < other.priority

    def is_condition_valid(self):
        """
        The condition for the event to be valid in order to be processed.
        By default the value returned is True
        """
        return True

    def generate_b_events(self, absolute_simulation_time):
        """
        Given an absolute_simulation_time, this method returns the set of BEvents
        triggered by this particular c_event.
        """
        pass



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from enum import Enum


class FailureType(Enum):
    DETECTABLE = "DETECTABLE"
    BLIND = "BLIND"

    def __str__(self):
        return str(self.name)

    def __eq__(self, other):
        return self.value == other.value


class Failure:
    __slots__ = 'type_of_failure'

    def __init__(self, type_of_failure: FailureType):
        self.type_of_failure = type_of_failure

    def __str__(self):
        return f"Failure: {self.type_of_failure}"

    def __eq__(self,other):
        return self.type_of_failure == other.type_of_failure


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import pandas

from availsim4core.src.analysis import Analysis
from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.exporter.result_exporter import ResultExporter
from availsim4core.src.results.results import Results


class XLSXResultExporterConnectivityMatrix(ResultExporter):
    @classmethod
    def export(cls,
               context: Context,
               output_folder: str,
               analysis: Analysis,
               results: Results,
               execution_time: float,
               unique_output_file_identifier_str: str):

        def connectivity_matrix_children_export(root: Component, level=0):
            """
            Function building a matrix with +1 recursively exploring the system in order to build a graphical representation of the system.
            This "extended" version of the tree exporter provides a detailed output. A simple version exists too.
            The function is called on the root component. Then if a component has children, it is called again for its
            children.
            """

            set_of_components = root.to_set()

            ret = ';'

            # header
            for icol, component_col in enumerate(set_of_components):
                ret += f"{component_col.name}_{component_col.local_id}_{component_col.global_id}; "
            ret += '\n'

            for irow, component_row in enumerate(set_of_components):

                # header row wise
                ret += f"{component_row.name}_{component_row.local_id}_{component_row.global_id}; "

                child_list = component_row.get_children()
                parent_list = component_row._parents

                for icol, component_col in enumerate(set_of_components):

                    if component_col in child_list:
                        ret+='1;'
                    elif component_col in parent_list:
                        ret+='-1;'
                    else:
                        ret+='0;'

                ret+='\n'

            return ret

        output_file = cls.OUTPUT_FILE_TEMPLATE.format(output_folder, analysis.id, unique_output_file_identifier_str)
        writer = ResultExporter.open_workbook(output_file)

        connectivity_matrix_string = connectivity_matrix_children_export(context.root_component)

        connectivity_matrix = pandas.DataFrame([x.split(';') for x in connectivity_matrix_string.split('\n')])

        connectivity_matrix.to_excel(writer, sheet_name='RESULTS_CONNECTIVITY_MATRIX', index=False)

        writer.save()


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

class Inspection:
    """
    Defines the failure attributes based on a probability law.
    Also defines the repair attributes related to this failure based on a probability law.
    # TODO : Refactoring of this class introducing dedicated classes for the failure / repair / inspection :
    cf https://gitlab.cern.ch/availsim4/availsim4core/-/issues/30
    """
    __slots__ = 'name', 'periodicity', 'duration'

    def __init__(self,
                 name,
                 inspection_periodicity: float,
                 inspection_duration: float):
        self.name = name
        self.periodicity = inspection_periodicity
        self.duration = inspection_duration

    def __hash__(self):
        return hash((type(self), self.name))

    def __str__(self):
        return f"{self.name} :: every {self.periodicity} during {self.duration}"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if not isinstance(other, Inspection):
            return NotImplemented
        return self.name == other.name



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import List

from availsim4core.src.context.context import Context
from availsim4core.src.results.results import Results


class SimulationType:
    MONTE_CARLO = "MONTE_CARLO"
    QUASI_MONTE_CARLO = "QUASI_MONTE_CARLO"
    SPLITTING_MONTE_CARLO = "SPLITTING_MONTE_CARLO"


class Simulation:
    """
    Defines a simulation that will be triggered by the Sensitivity Analysis.
    When a simulation is ran, it returns a ResultSimulation.
    """

    def run(self, context: Context) -> Results:
        """
        Run a simulation against a given Component tree by calling multiple Discrete event Simulations.
        """
        pass

    def get_list_of_diagnosis(self) -> List[str]:
        """
        Get a simulation's list of diagnosis for the exporter
        """
        pass



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging
from typing import List

from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.context.system.architecture_entry import ArchitectureEntry
from availsim4core.src.context.system.failure_factory import FailureFactory
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.context.system.failure_mode_assignments import FailureModeAssignments
from availsim4core.src.context.system.inspection import Inspection
from availsim4core.src.context.system.minimal_replaceable_unit import MinimalReplaceableUnitFactory
from availsim4core.src.context.system.phase_jump_trigger import PhaseJumpTriggerFactory
from availsim4core.src.context.system.probability_law.probability_law_factory import ProbabilityLawFactory
from availsim4core.src.context.system.rca_trigger import RootCauseAnalysisTriggerFactory
from availsim4core.src.context.system.system_template import SystemTemplate
from availsim4core.src.reader.xlsx.xlsx_reader import XLSXReader


class SystemTemplateSheet:
    ARCHITECTURE = "ARCHITECTURE"
    FAILURE_MODES = "FAILURE_MODES"
    FAILURE_MODE_ASSIGNMENTS = "FAILURE_MODE_ASSIGNMENTS"
    MINIMAL_REPLACEABLE_UNIT = "MRU"
    INSPECTIONS = "INSPECTIONS"
    PHASES = "PHASES"
    ROOT_CAUSE_ANALYSIS = "ROOT_CAUSE_ANALYSIS"
    PHASE_JUMP = "PHASE_JUMP"


class SystemTemplateArchitectureColumn:
    COMPONENT_NAME = "COMPONENT_NAME"
    COMPONENT_TYPE = "COMPONENT_TYPE"
    COMPONENT_NUMBER = "COMPONENT_NUMBER"
    CHILDREN_NAME = "CHILDREN_NAME"
    CHILDREN_LOGIC = "CHILDREN_LOGIC"
    IN_MRU = "IN_MRU"
    TRIGGER_MRU = "TRIGGER_MRU"


class SystemTemplateFailureModeAssignmentsColumn:
    COMPONENT_NAME = "COMPONENT_NAME"
    FAILURE_MODE_NAME = "FAILURE_MODE_NAME"


class SystemTemplateFailureModeColumn:
    FAILURE_MODE_NAME = "FAILURE_MODE_NAME"
    FAILURE_LAW = "FAILURE_LAW"
    FAILURE_PARAMETERS = "FAILURE_PARAMETERS"
    REPAIR_LAW = "REPAIR_LAW"
    REPAIR_PARAMETERS = "REPAIR_PARAMETERS"
    TYPE_OF_FAILURE = "TYPE_OF_FAILURE"
    HELD_BEFORE_REPAIR = "HELD_BEFORE_REPAIR"
    INSPECTION_NAME = "INSPECTION_NAME"
    PHASE_NAME = "PHASE_NAME"
    PHASE_NEXT_IF_FAILURE_NAME = "NEXT_PHASE_IF_FAILURE"
    PHASE_CHANGE_TRIGGER = "PHASE_CHANGE_TRIGGER"
    HELD_AFTER_REPAIR = "HELD_AFTER_REPAIR"


class SystemTemplateMinimalReplaceableUnitColumn:
    MINIMAL_REPLACEABLE_UNIT_NAME = "MRU_NAME"
    REPAIR_LAW = "MRU_LAW"
    REPAIR_PARAMETERS = "MRU_PARAMETERS"
    REPAIR_SCHEDULE = "MRU_SCHEDULE"
    LOWEST_COMMON_ANCESTOR = "LOWEST_COMMON_ANCESTOR_SCOPE"
    TRIGGERING_STATUS = "TRIGGERING_STATUS"


class SystemTemplateInspectionsColumn:
    INSPECTION_NAME = "INSPECTION_NAME"
    INSPECTION_PERIOD = "INSPECTION_PERIOD"
    INSPECTION_DURATION = "INSPECTION_DURATION"


class SystemTemplatePhasesColumn:
    PHASE_NAME = "PHASE_NAME"
    PHASE_LAW = "PHASE_LAW"
    PHASE_PARAMETERS = "PHASE_PARAMETERS"
    PHASE_NEXT = "NEXT_DEFAULT_PHASE"
    PHASE_FIRST = "FIRST_PHASE"
    PHASE_NEXT_IF_FAILURE = "NEXT_DEFAULT_PHASE_IF_FAILURE"


class SystemTemplateRootCauseAnalysisColumn:
    COMPONENT_NAME = "TRIGGERING_COMPONENT_NAME"
    COMPONENT_STATUS = "TRIGGERED_BY_COMPONENT_STATUS"
    PHASE = "TRIGGERED_IN_PHASE"


class SystemTemplatePhaseJumpColumn:
    COMPONENT_NAME = "TRIGGERING_COMPONENT_NAME"
    COMPONENT_STATUS = "TRIGGERED_BY_COMPONENT_STATUS"
    FROM_PHASE = "FROM_PHASE"
    TO_PHASE = "TO_PHASE"


class SystemTemplateField:
    NONE = "NONE"


class DuplicatedEntry(Exception):
    pass


class SystemTemplateReader(XLSXReader):

    @classmethod
    def generate_system_template(cls,
                                 system_file_path: str,
                                 custom_children_logic_path: str):
        """
        Given a file `system_file_path` which describe the system template.
        This methods generates the corresponding {SystemTemplate} object.
        :return SystemTemplate based on the given file path.
        """
        logging.debug(f"Reading System File = {system_file_path}")

        system_dictionary = cls.read(system_file_path)

        architecture_entry_list = cls.generate_architecture_entry_list(
            system_dictionary[SystemTemplateSheet.ARCHITECTURE])

        mru_list = cls.generate_mrus(system_dictionary[SystemTemplateSheet.MINIMAL_REPLACEABLE_UNIT])

        inspections_list = cls.generate_inspections(system_dictionary[SystemTemplateSheet.INSPECTIONS])

        phases_list = cls.generate_phases(system_dictionary[SystemTemplateSheet.PHASES])

        failure_modes_list = cls.generate_failure_modes(system_dictionary[SystemTemplateSheet.FAILURE_MODES],
                                                        inspections_list,
                                                        phases_list)

        failure_mode_assignments_list = cls.generate_failure_mode_assignments(
            system_dictionary[SystemTemplateSheet.FAILURE_MODE_ASSIGNMENTS],
            failure_modes_list)

        root_cause_analysis_triggers_list = cls.generate_root_cause_analysis_triggers(system_dictionary[SystemTemplateSheet.ROOT_CAUSE_ANALYSIS])

        phase_jump_triggers_list = cls.generate_phase_jump_triggers(system_dictionary[SystemTemplateSheet.PHASE_JUMP],
                                                                    phases_list)

        return SystemTemplate(architecture_entry_list,
                              failure_mode_assignments_list,
                              failure_modes_list,
                              mru_list,
                              inspections_list,
                              phases_list,
                              set(root_cause_analysis_triggers_list),
                              set(phase_jump_triggers_list),
                              custom_children_logic_path)

    @classmethod
    def generate_architecture_entry_list(cls, system_dictionary_architecture):
        """
        Extract from the given system_dictionary_architecture the list of the {ArchitectureEntry} of the global system.
        :param system_dictionary_architecture: the system_dictionary failure mode under the panda dictionary format.
        :return: List of {ArchitectureEntry} of the global system.
        """

        architecture_entry_list = []

        for row in system_dictionary_architecture.values():
            try:

                exception_message_hint = f"architecture row:{row}"

                component_name_str = cls.clean_cell_str_else_raise(row[SystemTemplateArchitectureColumn.COMPONENT_NAME],
                                                                   exception_message_hint=exception_message_hint)

                if component_name_str in [architecture_entry.component_name for architecture_entry in architecture_entry_list]:
                    message = f"The component name {component_name_str} is not unique in the list of architecture entries"
                    logging.exception(message)
                    raise DuplicatedEntry(message)
                
                component_type_str = cls.clean_cell_str_else_raise(row[SystemTemplateArchitectureColumn.COMPONENT_TYPE],
                                               exception_message_hint=exception_message_hint)

                component_number = row[SystemTemplateArchitectureColumn.COMPONENT_NUMBER]

                children_name_str = row[SystemTemplateArchitectureColumn.CHILDREN_NAME].upper()
                children_name_list = [] if children_name_str == SystemTemplateField.NONE \
                    else cls.extract_from_str_list(children_name_str, exception_message_hint=exception_message_hint)

                children_logic_str = cls.clean_cell_str_else_raise(row[SystemTemplateArchitectureColumn.CHILDREN_LOGIC],
                                                   exception_message_hint=exception_message_hint)
                children_logic_str = '' if children_logic_str == SystemTemplateField.NONE else children_logic_str

                in_mru_line = row[SystemTemplateArchitectureColumn.IN_MRU].upper()
                in_mru_str_list = [] if in_mru_line == SystemTemplateField.NONE \
                    else cls.extract_from_str_list(in_mru_line,exception_message_hint=exception_message_hint)

                trigger_mru_line = row[SystemTemplateArchitectureColumn.TRIGGER_MRU].upper()
                trigger_mru_str_list = [] if trigger_mru_line == SystemTemplateField.NONE \
                    else cls.extract_from_str_list(trigger_mru_line,exception_message_hint=exception_message_hint)

                architecture_entry_list.append(
                    ArchitectureEntry(
                        component_name_str,
                        component_type_str,
                        component_number,
                        children_name_list,
                        children_logic_str,
                        in_mru_str_list,
                        trigger_mru_str_list
                    )
                )
            except AttributeError:
                logging.info(f"Non-empty line with missing content present in the architecture sheet."
                             f"\nCheck row:{row}")

        logging.debug(f"Extracted from system file architecture_entry_list = {architecture_entry_list}")

        return architecture_entry_list

    @classmethod
    def generate_mrus(cls, system_dictionary_mru):
        """
        Extract from the given system_dictionary_mru the list of the MRUs of the global system.
        :param system_dictionary_mru: the system_dictionary mru under the panda dictionary format.
        see> SystemTemplate
        :return: List of Minimal Replaceable Unit of the global system.
        """

        mru_list = []

        for row in system_dictionary_mru.values():
            try:

                exception_message_hint = f"mru row:{row}"

                mru_name_str = cls.clean_cell_str_else_raise(
                    row[SystemTemplateMinimalReplaceableUnitColumn.MINIMAL_REPLACEABLE_UNIT_NAME],
                    exception_message_hint=exception_message_hint)

                if mru_name_str in [mru.name for mru in mru_list]:
                    message = f"The mru name {mru_name_str} is not unique in the list of mru entries"
                    logging.exception(message)
                    raise DuplicatedEntry(message)

                mru_repair_law_name = cls.clean_cell_str_else_raise(
                    row[SystemTemplateMinimalReplaceableUnitColumn.REPAIR_LAW],
                    exception_message_hint=exception_message_hint)

                mru_repair_parameters = row[SystemTemplateMinimalReplaceableUnitColumn.REPAIR_PARAMETERS]
                mru_repair_parameters = [mru_repair_parameters] if isinstance(mru_repair_parameters, (int, float)) \
                    else cls.extract_from_str_list(mru_repair_parameters,exception_message_hint=exception_message_hint)

                minimal_replaceable_repair_schedule = cls.clean_cell_str_else_raise(
                    row[SystemTemplateMinimalReplaceableUnitColumn.REPAIR_SCHEDULE],
                    exception_message_hint=exception_message_hint)

                lowest_common_parent_name_list = cls.extract_from_str_list(
                    row[SystemTemplateMinimalReplaceableUnitColumn.LOWEST_COMMON_ANCESTOR],
                    exception_message_hint=exception_message_hint)

                triggering_status_list_of_str = cls.extract_from_str_list(
                    row[SystemTemplateMinimalReplaceableUnitColumn.TRIGGERING_STATUS],
                    exception_message_hint=exception_message_hint)

                mru = MinimalReplaceableUnitFactory.build(mru_name_str,
                                                          mru_repair_law_name,
                                                          mru_repair_parameters,
                                                          minimal_replaceable_repair_schedule,
                                                          triggering_status_list_of_str,
                                                          lowest_common_parent_name_list)

                mru_list.extend(mru)
            except AttributeError:
                logging.info("Non-empty line with missing content present in the MRU sheet."
                             f"\nCheck row:{row}")

        logging.debug(f"Extracted from system file mru_list = {mru_list}")

        return mru_list

    @classmethod
    def generate_inspections(cls, system_dictionary_inspections):
        """
        Extract from the given system_dictionary_inspections the list of the inspection of the global system.
        :param system_dictionary_inspections: the system_dictionary inspections under the panda dictionary format.
        see> SystemTemplate
        :return: List of inspections of the global system.
        """

        inspections_list = []

        for row in system_dictionary_inspections.values():
            try:

                exception_message_hint = f"inspection row: {row}"

                inspection_name_str = cls.clean_cell_str_else_raise(row[SystemTemplateInspectionsColumn.INSPECTION_NAME],
                                                exception_message_hint=exception_message_hint)

                if inspection_name_str in [inspection.name for inspection in inspections_list]:
                    message = f"The inspection {inspection_name_str} is not unique in the list of inspections"
                    logging.exception(message)
                    raise DuplicatedEntry(message)

                inspection_period = row[SystemTemplateInspectionsColumn.INSPECTION_PERIOD]

                inspection_duration = row[SystemTemplateInspectionsColumn.INSPECTION_DURATION]

                inspections_list.append(Inspection(inspection_name_str,
                                                   inspection_period,
                                                   inspection_duration))
            except AttributeError:
                logging.info("Non-empty line with missing content present in the Inspections sheet."
                             f"\nCheck row:{row}")

        logging.debug(f"Extracted from system file inspections_list = {inspections_list}")

        return inspections_list

    @classmethod
    def generate_phases(cls, system_dictionary_phases):
        """
        Extract from the given `system_dictionary_phase` the list of the {Phase} of the global system.
        :param system_dictionary_phases: the system_dictionary_phases under the panda dictionary format.
        :return: List of {Phase} of the global system.
        """

        phases_list = []
        dict_phase_next = {}
        dict_next_phase_if_failure = {}

        for row in system_dictionary_phases.values():
            try:

                exception_message_hint = f"phases row: {row}"

                phase_name_str = cls.clean_cell_str_else_raise(row[SystemTemplatePhasesColumn.PHASE_NAME],exception_message_hint=exception_message_hint)

                if phase_name_str in [phase.name for phase in phases_list]:
                    message = f"The phase {phase_name_str} is not unique in the list of phases"
                    logging.exception(message)
                    raise DuplicatedEntry(message)

                phase_law_str = cls.clean_cell_str_else_raise(row[SystemTemplatePhasesColumn.PHASE_LAW],exception_message_hint=exception_message_hint)

                phase_parameters = row[SystemTemplatePhasesColumn.PHASE_PARAMETERS]
                phase_parameters = [phase_parameters] if isinstance(phase_parameters, (int, float)) \
                    else cls.extract_from_str_list(phase_parameters,exception_message_hint=exception_message_hint)

                phase_next_name_str = cls.clean_cell_str_else_raise(row[SystemTemplatePhasesColumn.PHASE_NEXT],exception_message_hint=exception_message_hint)

                phase_first = bool(row[SystemTemplatePhasesColumn.PHASE_FIRST])

                next_phase_if_failure_name_str = cls.clean_cell_str_else_raise(row[SystemTemplatePhasesColumn.PHASE_NEXT_IF_FAILURE],
                                                               exception_message_hint=exception_message_hint)

                phase_law = ProbabilityLawFactory.build(phase_law_str,
                                                        phase_parameters)

                not_linked_phase = Phase(phase_name_str,
                                         phase_law,
                                         phase_first)

                dict_phase_next[not_linked_phase] = phase_next_name_str

                dict_next_phase_if_failure[not_linked_phase] = next_phase_if_failure_name_str

                phases_list.append(not_linked_phase)
            except AttributeError:
                logging.info("Non-empty line with missing content present in the Phases sheet."
                             f"\nCheck row:{row}")

        for not_linked_phase in phases_list:
            for phase in phases_list:
                if phase.name == dict_phase_next[not_linked_phase]:
                    not_linked_phase.set_next_phase(phase)
                if phase.name == dict_next_phase_if_failure[not_linked_phase]:
                    not_linked_phase.set_next_phase_if_failure(phase)

        if not phases_list:
            phases_list = [PhaseManager.NO_PHASE]  # Default is NO_PHASE

        logging.debug(f"Extracted from system file phases_list = {phases_list}")

        return phases_list

    @classmethod
    def generate_failure_modes(cls,
                               system_dictionary_failure_mode,
                               inspections_list,
                               phase_list):
        """
        Extract from the given system_dictionary_failure_mode the list of the FailureModes of the global system.
        :param system_dictionary_failure_mode: the system_dictionary failure mode under the panda dictionary format.
        see> SystemTemplate
        :param inspections_list: the list of inspections objects
        :param phase_list: List of phases {Phase}
        :return: List of FailureMode of the global system.
        """

        failure_modes_list = []

        for row in system_dictionary_failure_mode.values():
            try:

                exception_message_hint = f"failure mode row: {row}"

                failure_mode_name_str = cls.clean_cell_str_else_raise(row[SystemTemplateFailureModeColumn.FAILURE_MODE_NAME],
                                                      exception_message_hint=exception_message_hint)

                if failure_mode_name_str in [failure_mode.name for failure_mode in failure_modes_list]:
                    message = f"The failure mode {failure_mode_name_str} is not unique in the list of failure modes"
                    logging.exception(message)
                    raise DuplicatedEntry(message)

                failure_mode_failure_law = cls.clean_cell_str_else_raise(row[SystemTemplateFailureModeColumn.FAILURE_LAW],
                                                         exception_message_hint=exception_message_hint)

                failure_mode_failure_parameters = row[SystemTemplateFailureModeColumn.FAILURE_PARAMETERS]
                failure_mode_failure_parameters = [failure_mode_failure_parameters] \
                    if isinstance(failure_mode_failure_parameters, (int, float)) \
                    else cls.extract_from_str_list(failure_mode_failure_parameters, exception_message_hint=exception_message_hint)

                failure_mode_repair_law = cls.clean_cell_str_else_raise(row[SystemTemplateFailureModeColumn.REPAIR_LAW],
                                                        exception_message_hint=exception_message_hint)

                failure_mode_repair_parameters = row[SystemTemplateFailureModeColumn.REPAIR_PARAMETERS]
                failure_mode_repair_parameters = [failure_mode_repair_parameters] \
                    if isinstance(failure_mode_repair_parameters, (int, float)) \
                    else cls.extract_from_str_list(failure_mode_repair_parameters, exception_message_hint=exception_message_hint)

                failure_mode_type_of_failure = cls.clean_cell_str_else_raise(row[SystemTemplateFailureModeColumn.TYPE_OF_FAILURE],
                                                             exception_message_hint=exception_message_hint)

                failure_mode_failure = FailureFactory.build(failure_mode_type_of_failure)

                failure_mode_phase_str = row[SystemTemplateFailureModeColumn.PHASE_NAME]
                failure_mode_phase_str = cls.extract_from_str_list(failure_mode_phase_str, exception_message_hint=exception_message_hint)

                failure_mode_next_phase_if_failure_name_str = cls.clean_cell_str_else_raise(
                    row[SystemTemplateFailureModeColumn.PHASE_NEXT_IF_FAILURE_NAME],
                    exception_message_hint=exception_message_hint)

                failure_mode_held_before_repair_str = cls.extract_from_str_list(
                    row[SystemTemplateFailureModeColumn.HELD_BEFORE_REPAIR],
                    exception_message_hint=exception_message_hint)

                failure_mode_phase_change_trigger = cls.clean_cell_str_else_raise(
                    row[SystemTemplateFailureModeColumn.PHASE_CHANGE_TRIGGER],
                    exception_message_hint=exception_message_hint)

                failure_mode_held_after_repair_str = cls.extract_from_str_list(
                    row[SystemTemplateFailureModeColumn.HELD_AFTER_REPAIR],
                    exception_message_hint=exception_message_hint)

                failure_law = ProbabilityLawFactory.build(failure_mode_failure_law,
                                                          failure_mode_failure_parameters)

                repair_law = ProbabilityLawFactory.build(failure_mode_repair_law,
                                                         failure_mode_repair_parameters)

                failure_mode_inspection_name_line = row[SystemTemplateFailureModeColumn.INSPECTION_NAME].upper()
                failure_mode_inspection = next(
                    (inspection
                    for inspection in inspections_list
                    if inspection.name == failure_mode_inspection_name_line),
                    None)

                failure_mode_phase_set = {phase
                                        for phase in phase_list
                                        if phase.name in failure_mode_phase_str}

                failure_mode_next_phase_if_failure = None
                if failure_mode_next_phase_if_failure_name_str != SystemTemplateField.NONE:
                    for phase in phase_list:
                        if phase.name == failure_mode_next_phase_if_failure_name_str:
                            failure_mode_next_phase_if_failure = phase

                if failure_mode_held_after_repair_str[0] == "HELD_FOREVER":
                    failure_mode_held_after_repair_phase_set= {PhaseManager.HELD_FOREVER}  # Special phase used to never release some held components
                elif failure_mode_held_after_repair_str[0] == "NEVER_HELD":
                    failure_mode_held_after_repair_phase_set = set(phase_list)
                else:
                    failure_mode_held_after_repair_phase_set = {phase
                                                for phase in phase_list
                                                if phase.name in failure_mode_held_after_repair_str}

                if failure_mode_held_before_repair_str[0] == "HELD_FOREVER":
                    failure_mode_held_before_repair_phase_set= {PhaseManager.HELD_FOREVER}  # Special phase used to never release some held components
                elif failure_mode_held_before_repair_str[0] == "NEVER_HELD":
                    failure_mode_held_before_repair_phase_set = set(phase_list)
                else:
                    failure_mode_held_before_repair_phase_set = {phase
                                                for phase in phase_list
                                                if phase.name in failure_mode_held_before_repair_str}

                failure_modes_list.append(FailureMode(failure_mode_name_str,
                                                      failure_law,
                                                      repair_law,
                                                      failure_mode_failure,
                                                      failure_mode_held_before_repair_phase_set,
                                                      failure_mode_inspection,
                                                      failure_mode_phase_set,
                                                      failure_mode_next_phase_if_failure,
                                                      failure_mode_phase_change_trigger,
                                                      failure_mode_held_after_repair_phase_set))
            except AttributeError:
                logging.info("Non-empty line with missing content present in the failure_modes sheet."
                             f"\nCheck row:{row}")

        logging.debug(f"Extracted from system file failure_modes_list = {failure_modes_list}")
        return failure_modes_list

    @classmethod
    def generate_failure_mode_assignments(cls,
                                          system_dictionary_assignments,
                                          failure_mode_list):
        """
        Generates failure mode assignments containing failure mode components by matching failure modes from the given
        `system_dictionary_assignments` with failure modes in the `failure modes list`.
        :param system_dictionary_assignments: The dictionary of failure assignments under the panda dataframe format.
        see> SystemTemplate.
        :param failure_mode_list: The List of Failure Mode to be associated to each of the FailureModeAssignment.
        :return: The complete list of the FailureModeAssignment for the system.
        """
        failure_mode_assignments_list = []
        for row in system_dictionary_assignments.values():
            try:

                exception_message_hint = f"failure mode assignements row: {row}"

                failure_mode_assignments_component_name_str = cls.clean_cell_str_else_raise(
                    row[SystemTemplateFailureModeAssignmentsColumn.COMPONENT_NAME],
                    exception_message_hint=exception_message_hint)
                if failure_mode_assignments_component_name_str in [failure_mode_assignment.component_name
                                                                   for failure_mode_assignment
                                                                   in failure_mode_assignments_list]:
                    message = f"The component name {failure_mode_assignments_component_name_str} is not unique in the list of failure mode assignement names"
                    logging.exception(message)
                    raise DuplicatedEntry(message)

                failure_mode_assignments_failure_mode_name_str = cls.clean_cell_str_else_raise(
                    row[SystemTemplateFailureModeAssignmentsColumn.FAILURE_MODE_NAME],
                    exception_message_hint=exception_message_hint)

                failure_mode = [failure_mode for failure_mode in failure_mode_list
                                if failure_mode.name == failure_mode_assignments_failure_mode_name_str][0]
                failure_mode_assignments_list.append(FailureModeAssignments(failure_mode_assignments_component_name_str,
                                                                            failure_mode))
            except AttributeError:
                logging.info("Non-empty line with missing content present in the failure_mode_assignments sheet."
                             f"\nCheck row:{row}")

        logging.debug(f"Extracted from system file failure_mode_assignments_list = {failure_mode_assignments_list}")
        return failure_mode_assignments_list

    @classmethod
    def generate_root_cause_analysis_triggers(cls, system_dictionary_root_cause_analysis):
        """
        Generates a list of components and their statuses which trigger Root Cause Analysis dumps.
        :param system_dictionary_root_cause_analysis:
        :return: The complete list of RCA triggers for the system.
        """

        root_cause_analysis_triggers_list = []

        for row in system_dictionary_root_cause_analysis.values():
            try:

                exception_message_hint = f"root cause analysis row: {row}"

                root_cause_analysis_trigger_component_name = cls.clean_cell_str_else_raise(
                    row[SystemTemplateRootCauseAnalysisColumn.COMPONENT_NAME],
                    exception_message_hint=exception_message_hint)
                root_cause_analysis_trigger_component_status_list = cls.extract_from_str_list(
                    row[SystemTemplateRootCauseAnalysisColumn.COMPONENT_STATUS],
                    exception_message_hint=exception_message_hint)
                root_cause_analysis_trigger_phase_list = cls.extract_from_str_list(
                    row[SystemTemplateRootCauseAnalysisColumn.PHASE],
                    exception_message_hint=exception_message_hint)

                rca = RootCauseAnalysisTriggerFactory.build(
                    root_cause_analysis_trigger_component_name,
                    root_cause_analysis_trigger_component_status_list,
                    root_cause_analysis_trigger_phase_list
                )

                root_cause_analysis_triggers_list.extend(rca)
            except AttributeError:
                logging.info("Non-empty line with missing content present in the root_cause_analysis sheet."
                             f"\nCheck row:{row}")

        logging.debug(f"Extracted from system file root_cause_analysis_list = {root_cause_analysis_triggers_list}")
        return root_cause_analysis_triggers_list

    @classmethod
    def generate_phase_jump_triggers(cls,
                                     system_dictionary_phase_jump,
                                     phase_list: List[Phase]):
        """
        Generates a list of (component, status, in_phase, jump_to_phase) which trigger phase jumps.
        :param system_dictionary_phase_jump:
        :return: The complete list of phase jump triggers.
        """

        phase_jump_triggers_list = []

        for row in system_dictionary_phase_jump.values():
            try:

                exception_message_hint = f"phase jump row: {row}"

                phase_jump_trigger_component_name = cls.clean_cell_str_else_raise(
                    row[SystemTemplatePhaseJumpColumn.COMPONENT_NAME],
                    exception_message_hint=exception_message_hint)
                phase_jump_trigger_component_status_list = cls.extract_from_str_list(
                    row[SystemTemplatePhaseJumpColumn.COMPONENT_STATUS],
                    exception_message_hint=exception_message_hint)
                phase_jump_trigger_from_phase_list = cls.extract_from_str_list(
                    row[SystemTemplatePhaseJumpColumn.FROM_PHASE],
                    exception_message_hint=exception_message_hint)
                phase_jump_trigger_to_phase = cls.clean_cell_str_else_raise(
                    row[SystemTemplatePhaseJumpColumn.TO_PHASE],
                    exception_message_hint=exception_message_hint)

                phase_jump = PhaseJumpTriggerFactory.build(
                    phase_jump_trigger_component_name,
                    phase_jump_trigger_component_status_list,
                    phase_jump_trigger_from_phase_list,
                    phase_jump_trigger_to_phase,
                    phase_list
                )

                phase_jump_triggers_list.extend(phase_jump)
            except AttributeError:
                logging.info("Non-empty line with missing content present in the phase_jump sheet."
                             f"\nCheck row:{row}")

        logging.debug(f"Extracted from system file phase_jump_triggers_list = {phase_jump_triggers_list}")
        return phase_jump_triggers_list


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

class ArchitectureEntry:

    def __init__(self,
                 component_name,
                 component_type_str,
                 component_number,
                 children_name_list,
                 children_logic_str,
                 in_mru_str_list,
                 trigger_mru_str_list):
        self.component_name = component_name
        self.component_type_str = component_type_str
        self.component_number = component_number
        self.children_name_list = children_name_list
        self.children_logic_str = children_logic_str
        self.in_mru_str_list = in_mru_str_list
        self.trigger_mru_str_list = trigger_mru_str_list

    def __str__(self):
        return (f"name: {self.component_name}; "
                f"type: {self.component_type_str}; "
                f"number: {self.component_number}; "
                f"children: {self.children_name_list}; "
                f"logic: {self.children_logic_str}; "
                f"in mru: {self.in_mru_str_list}; "
                f"trigger mru: {self.trigger_mru_str_list}")

    def __eq__(self, other):
        return self.component_name == other.component_name \
               and self.component_type_str == other.component_type_str \
               and self.component_number == other.component_number \
               and self.children_name_list == other.children_name_list \
               and self.children_logic_str == other.children_logic_str \
               and self.in_mru_str_list == other.in_mru_str_list \
               and self.trigger_mru_str_list == other.trigger_mru_str_list

    def __repr__(self):
        return str(self)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import TYPE_CHECKING, Set, List, Tuple

from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.discrete_event_simulation.event.b_event.b_event import BEventPriority

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.basic import Basic

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.discrete_event_simulation.event.b_event.basic_b_event import BasicBEvent
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEvent, CEventPriority
from availsim4core.src.discrete_event_simulation.event.c_event.failure_event.order_failure_event import \
    OrderFailureEvent
from availsim4core.src.discrete_event_simulation.event.event import Event


class EndRepairingEvent(BasicBEvent):
    """
    Defines a Repair event.
    Attributes:
        - absolute_occurrence_time: date when the repair event will be started.
        - context:
        - basic: associate basic to repair.
        - event: event which triggered this repair event.
    """
    __slots__ = 'event', 'failure_mode'

    def __init__(self,
                 absolute_occurrence_time: float,
                 context: Context,
                 basic: Basic,
                 event: Event,
                 failure_mode: FailureMode):
        super().__init__(absolute_occurrence_time, context, basic,
                         BEventPriority.END_REPAIRING_EVENT)
        self.event = event
        self.failure_mode = failure_mode

    def __hash__(self):
        return hash((type(self), self.absolute_occurrence_time, self.priority, self.basic, self.failure_mode,
                     self.event))

    def __eq__(self, other):
        if not isinstance(other, EndRepairingEvent):
            return NotImplemented
        return super().__eq__(other) and \
               self.event == other.event and \
               self.failure_mode == other.failure_mode

    def __str__(self):
        return (f"type ::{type(self)} "
                f"basic ::{self.basic.name}, "
                f"lid ::{self.basic.local_id}, "
                f"gid ::{self.basic.global_id}, "
                f"at t ::{self.absolute_occurrence_time}, "
                f"FailureMode ::{self.failure_mode} \n")

    def execute(self):

        if self.context.phase_manager.current_phase not in self.failure_mode.held_after_repair_phase_set:
            return self.basic.update_status(Status.HELD,
                                        self.absolute_occurrence_time,
                                        self.context.phase_manager.current_phase,
                                        f"REPAIRED and set to HELD - component {self.basic.name}_{self.basic.local_id}_{self.basic.global_id}",
                                        self.context)
        else:
            return self.basic.update_status(Status.RUNNING,
                                            self.absolute_occurrence_time,
                                            self.context.phase_manager.current_phase,
                                            f"REPAIRED - component {self.basic.name}_{self.basic.local_id}_{self.basic.global_id}",
                                            self.context)

    def generate_c_event(self, **kwargs) -> Set[CEvent]:

        c_event_set = set()

        order_failure_event = OrderFailureEvent(
            priority=CEventPriority.ORDER_FAILURE_EVENT,
            context=self.context,
            component=self.basic,
            event=self.event,
            failure_mode=self.failure_mode
        )

        if self.failure_mode.held_after_repair_phase_set != {PhaseManager.HELD_FOREVER}:
            if self.context.phase_manager.current_phase not in self.failure_mode.held_after_repair_phase_set:
                from availsim4core.src.discrete_event_simulation.event.c_event.held_event.order_end_holding_event import \
                    OrderEndHoldingEvent
                c_event_set.add( OrderEndHoldingEvent(
                    priority=CEventPriority.ORDER_END_HOLDING_EVENT,
                    context=self.context,
                    component=self.basic,
                    event=self.event,
                    failure_mode=self.failure_mode,
                    held_event=order_failure_event,
                    held_until_phase_set=self.failure_mode.held_after_repair_phase_set
                ))
            else:
                c_event_set.add(order_failure_event)

        if self.failure_mode.phase_change_trigger in ["AFTER_REPAIR"]:

            priority = CEventPriority.ORDER_NEXT_PHASE_EVENT
            if self.failure_mode.failure_mode_next_phase_if_failure is not None:
                priority = CEventPriority.ORDER_NEXT_PHASE_EVENT_IF_SPECIFIC_FAILURE

            from availsim4core.src.discrete_event_simulation.event.c_event.phase_event.order_next_phase_if_failure_event import \
                OrderNextPhaseIfFailureEvent
            c_event_set.add(OrderNextPhaseIfFailureEvent(priority, self.context, self.failure_mode))

        return c_event_set

    def update_b_event_collection(self,
                                event_set: Set,
                                types_of_event_to_clean: List[type]) -> Tuple[Set, Set]:
        """
        The input event_set is cleaned from all the events of types provided by types_of_event_to_clean.
        Returns a clean set of Events and a set containing the removed events.
        """

        if self.failure_mode.phase_change_trigger in ["AFTER_REPAIR"]:
            event_to_remove = {event
                               for event in event_set
                               if type(event) in types_of_event_to_clean}

            return event_set - event_to_remove, event_to_remove

        return event_set, set()

    def update_c_event_collection(self,
                                event_set: Set,
                                types_of_event_to_clean: List[type]) -> Tuple[Set, Set]:
        """
        The input event_set is cleaned from all the events of types provided by types_of_event_to_clean.
        Returns a clean set of Events and a set containing the removed events.
        """

        if self.failure_mode.phase_change_trigger in ["AFTER_REPAIR"]:
            event_to_remove = {event
                               for event in event_set
                               if type(event) in types_of_event_to_clean}

            return event_set - event_to_remove, event_to_remove

        return event_set, set()

    @staticmethod
    def _b_events_to_be_cleaned() -> List:
        from availsim4core.src.discrete_event_simulation.event.b_event.phase_event.next_phase_event import \
            NextPhaseEvent
        from availsim4core.src.discrete_event_simulation.event.b_event.phase_event.next_phase_if_failure_event import \
            NextPhaseIfFailureEvent
        from availsim4core.src.discrete_event_simulation.event.b_event.phase_event.jump_phase_event import \
            JumpPhaseEvent
        return [NextPhaseEvent, NextPhaseIfFailureEvent, JumpPhaseEvent]

    @staticmethod
    def _c_events_to_be_cleaned() -> List:
        from availsim4core.src.discrete_event_simulation.event.c_event.phase_event.order_next_phase_event import \
            OrderNextPhaseEvent
        from availsim4core.src.discrete_event_simulation.event.c_event.phase_event.order_next_phase_if_failure_event import \
            OrderNextPhaseIfFailureEvent
        return [OrderNextPhaseEvent,OrderNextPhaseIfFailureEvent]


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import importlib
import logging
import os

from availsim4core.src.context.system.children_logic.and_ import And
from availsim4core.src.context.system.children_logic.children_logic import ChildrenLogic
from availsim4core.src.context.system.children_logic.oo import Oo
from availsim4core.src.context.system.children_logic.required_component import RequiredComponent
from availsim4core.src.context.system.children_logic.tolerated_fault import ToleratedFault


class ChildrenLogicFactoryError(Exception):
    pass


class ChildrenLogicFactory:

    @staticmethod
    def moduleNameFromPath(custom_children_logic_path: str) -> str:
        """
        function getting the name of the module out of the path of the custom children logic file
        :param custom_children_logic_path: path of the custom children logic file
        :return: name of the module
        """
        return '.'+os.path.split(custom_children_logic_path)[-1].split(".")[0]

    @staticmethod
    def packageNameFromPath(custom_children_logic_path: str) -> str:
        """
        function getting the name of the package out of the path of the custom children logic file
        :param custom_children_logic_path: path of the custom children logic file
        :return: path of the package
        """
        name = ChildrenLogicFactory.moduleNameFromPath(custom_children_logic_path)
        return ''.join(custom_children_logic_path.replace('.', '').replace('/', '.').rsplit(name, 1)[0])

    @staticmethod
    def build(children_logic_str: str, custom_children_logic_path: str) -> ChildrenLogic:
        """
        Given a string children logic word, this method instantiates the corresponding children_logic class.
        :param children_logic_str: word to analyse from (not case sensitive.)
        :param custom_children_logic_path: string defining the path toward an optional python file containing custom
        :return: the children_logic instance corresponding to the given string.
        """

        if children_logic_str == "AND":
            return And()
        elif "OO" in children_logic_str:
            minimum_number_of_required_component = int(children_logic_str.split('OO')[0])
            total_number_of_component = int(children_logic_str.split('OO')[1])
            return Oo(minimum_number_of_required_component, total_number_of_component)
        elif "TF" in children_logic_str:
            fault_tolerance = int(children_logic_str.split('TF')[0])
            return ToleratedFault(fault_tolerance)
        elif "RC" in children_logic_str:
            required_component = int(children_logic_str.split('RC')[0])
            return RequiredComponent(required_component)
        elif custom_children_logic_path:
            name = ChildrenLogicFactory.moduleNameFromPath(custom_children_logic_path)
            package = ChildrenLogicFactory.packageNameFromPath(custom_children_logic_path)
            mod = importlib.import_module(name, package)
            custom_children_logic = getattr(mod, children_logic_str)
            return custom_children_logic()
        else:
            exception_message = f"{children_logic_str} not supported type of children logic."
            logging.exception(exception_message)
            raise ChildrenLogicFactoryError(exception_message)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.context import Context
from availsim4core.src.discrete_event_simulation.event.b_event.b_event import BEventPriority
from availsim4core.src.discrete_event_simulation.event.b_event.phase_event.next_phase_event import NextPhaseEvent
from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.context.system.failure_mode import FailureMode


class NextPhaseIfFailureEvent(NextPhaseEvent):
    """
    event class that is in charge of switching Phases when a failure occurs.
    """
    __slots__ = 'failure_mode'

    def __init__(self,
                 absolute_occurrence_time: float,
                 context: Context,
                 failure_mode: FailureMode):

        if failure_mode.failure_mode_next_phase_if_failure is None:
            priority = BEventPriority.NEXT_PHASE_IF_FAILURE_EVENT
        else:
            priority = BEventPriority.NEXT_PHASE_IF_SPECIFIC_FAILURE_EVENT

        super().__init__(absolute_occurrence_time,
                         context,
                         priority)
        self.failure_mode = failure_mode

    def __hash__(self):
        return hash((type(self), self.absolute_occurrence_time, self.priority, self.failure_mode))

    def __eq__(self, other):
        if not isinstance(other, NextPhaseIfFailureEvent):
            return NotImplemented
        return super().__eq__(other) and \
               self.failure_mode == other.failure_mode

    def __str__(self):
        return (f"type ::{type(self)} "
                f"due to ::{self.failure_mode} "
                f"at t ::{self.absolute_occurrence_time}\n")

    def execute(self):
        if self.failure_mode.failure_mode_next_phase_if_failure is None:
            records = [self.context.phase_manager.go_to_phase(
                self.absolute_occurrence_time,
                self.context.phase_manager.current_phase.next_phase_if_failure,
                PhaseManager.DEFAULT_PHASE_IF_FAILURE_DESCRIPTION)]
        else:
            records = [self.context.phase_manager.go_to_phase(
                self.absolute_occurrence_time,
                self.failure_mode.failure_mode_next_phase_if_failure,
                PhaseManager.SPECIFIC_PHASE_IF_FAILURE_DESCRIPTION)]
        records.extend(self.context.add_record_of_each_component(
            self.absolute_occurrence_time,
            self.context.phase_manager.UPDATE_FOLLOWING_A_PHASE_CHANGE))
        return records



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.sensitivity_analysis.exploration_strategy import Inner, Outer
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination_strategy import \
    SystemModifierCombinationStrategy


class test_SystemModifierCombinationStrategy(unittest.TestCase):

    def test_filter_exploration__inner(self):
        exploration_strategy_list = [Inner("inner", 0),
                                     Outer("outer", 1)]

        expected_result = [Inner("inner", 0)]

        result = SystemModifierCombinationStrategy.filter_exploration(exploration_strategy_list, Inner)

        self.assertEqual(expected_result, result)

    def test_filter_exploration__outer(self):
        exploration_strategy_list = [Inner("inner", 0),
                                     Outer("outer", 1)]

        expected_result = [Outer("outer", 1)]

        result = SystemModifierCombinationStrategy.filter_exploration(exploration_strategy_list, Outer)

        self.assertEqual(expected_result, result)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import TYPE_CHECKING, Set
import logging
import sys

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.basic import Component

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.minimal_replaceable_unit import MinimalReplaceableUnit
from availsim4core.src.context.system.system_utils import SystemUtils
from availsim4core.src.discrete_event_simulation.event.b_event.b_event import BEvent
from availsim4core.src.discrete_event_simulation.event.b_event. \
    repair_event.minimal_replaceable_unit_end_repairing_event import MinimalReplaceableUnitEndRepairingEvent
from availsim4core.src.discrete_event_simulation.event.b_event.repair_event. \
    minimal_replaceable_unit_start_repairing_event import MinimalReplaceableUnitStartRepairingEvent
from availsim4core.src.discrete_event_simulation.event.c_event.component_c_event import ComponentCEvent


class MinimalReplaceableUnitOrderRepairEvent(ComponentCEvent):
    """
   MinimalReplaceableUnit are a mechanism used to repair a group of components when one or a group of specific monitored
   components fail. This class generates the C event used in that mechanism
   """
    __slots__ = 'mru'

    def __init__(self,
                 priority: int,
                 context: Context,
                 component: Component,
                 mru: MinimalReplaceableUnit):
        super().__init__(priority, context, component)
        self.mru = mru

    def __eq__(self, other):
        if not isinstance(other, MinimalReplaceableUnitOrderRepairEvent):
            return NotImplemented
        return super().__eq__(other) and \
               self.mru == other.mru

    def __hash__(self):
        return hash((type(self), self.priority, self.component, self.mru))

    def __str__(self):
        return f"MinimalReplaceableUnitOrderRepairEvent:: " \
               f"priority:{self.priority} - " \
               f"component:{self.component.name} - " \
               f"mru:{self.mru.name}"

    def is_condition_valid(self):
        return self.component.status == self.mru.status

    def _create_mru_b_event(self,
                            absolute_simulation_time,
                            component):
        start_repairing_event = \
            MinimalReplaceableUnitStartRepairingEvent(absolute_occurrence_time=absolute_simulation_time,
                                                      context=self.context,
                                                      basic=component,
                                                      event=self)

        end_repairing_event = \
            MinimalReplaceableUnitEndRepairingEvent(
                absolute_occurrence_time=absolute_simulation_time + self.mru.repair_law.get_random_value(),
                context=self.context,
                basic=component,
                event=self,
                mru_trigger=self.mru)

        return {start_repairing_event, end_repairing_event}

    def get_b_events_from_mru_target(self,
                                     absolute_simulation_time) -> Set[BEvent]:
        """
        Finds the lowest common ancestor component, then for each of its children associated to the mru,
        generates the corresponding b_event.
        """

        # get up to the lowest common ancestor
        scope_common_ancestor_list = [SystemUtils.find_first_lowest_level_ancestor_with_name(self.component, scope_ancestor) 
                                                                                       for scope_ancestor in self.mru.scope_common_ancestor]
        try:
            scope_common_ancestor = next(ancestor for ancestor in scope_common_ancestor_list if ancestor is not None)
        except StopIteration as e:
            message = f"No component defining the size of MRU scope ({self.mru.scope_common_ancestor}) was located in the list of ancestors of the component {self.component}"
            logging.error(message)
            sys.exit(1)
            # TODO: SANITY CHECK: catch this error earlier in the process

        # get list of components subject to the trigger
        list_of_components_matching_mru = self.get_list_of_components_subject_to_the_trigger(
            scope_common_ancestor)

        return {b_event
                for mru_component in list_of_components_matching_mru
                for b_event in self._create_mru_b_event(absolute_simulation_time,
                                                        mru_component)}

    def get_list_of_components_subject_to_the_trigger(self, component):
        """
        For a given component, this method retrieves the associated matching minimal replaceable target units.
        """
        if not component.get_children():
            return [component
                    for mru_target in component.list_of_mru_group
                    if self.mru == mru_target]

        return [component_trigger
                for child in component.get_children()
                for component_trigger in self.get_list_of_components_subject_to_the_trigger(child)]

    def generate_b_events(self, absolute_simulation_time):
        return self.get_b_events_from_mru_target(absolute_simulation_time)
